"""Orbit Interlock devices."""

import numpy as _np

from ..clientconfigdb import ConfigDBClient
from ..search import BPMSearch as _BPMSearch, OrbIntlkSearch as _OrbIntlkSearch
from ..sofb.utils import si_calculate_bump as _si_calculate_bump
from ..util import ClassProperty as _classproperty
from .device import Device as _Device, DeviceSet as _DeviceSet


class BaseOrbitIntlk:
    """Base BPM Orbit Interlock Info."""

    CONV_NM2UM = 1e-3  # [nm] --> [um]

    UP_2_DOWN = {'M1': 'M2', 'C1-1': 'C1-2', 'C2': 'C3-1', 'C3-2': 'C4'}
    DOWN_2_UP = {val: key for key, val in UP_2_DOWN.items()}

    __BPM_NAMES = None
    __BPM_NICKNAMES = None
    __BPM_POS = None
    __BPM_DOWNSTREAM = None
    __BPM_UPSTREAM = None

    @_classproperty
    def BPM_NAMES(cls):
        """BPM names."""
        if cls.__BPM_NAMES is not None:
            return cls.__BPM_NAMES

        cls.__BPM_NAMES = _BPMSearch.get_names({'sec': 'SI', 'dev': 'BPM'})
        cls.__BPM_NICKNAMES = _BPMSearch.get_nicknames(cls.__BPM_NAMES)
        cls.__BPM_POS = _BPMSearch.get_positions(cls.__BPM_NAMES)

        def _parse_nick_down_up(name, nick):
            sub, idx, *_ = nick.split('-') + ['']
            return name.substitute(sub=name.sub[:2] + sub, idx=idx)

        downnames, upnames = [], []
        for bpm in cls.__BPM_NAMES:
            nick = bpm.sub[2:] + ('-' + bpm.idx if bpm.idx else '')
            if nick in BaseOrbitIntlk.DOWN_2_UP:
                down = bpm
                upnick = BaseOrbitIntlk.DOWN_2_UP[nick]
                upn = _parse_nick_down_up(bpm, upnick)
            elif nick in BaseOrbitIntlk.UP_2_DOWN:
                upn = bpm
                downnick = BaseOrbitIntlk.UP_2_DOWN[nick]
                down = _parse_nick_down_up(bpm, downnick)
            else:
                down, upn = '', ''
            downnames.append(down)
            upnames.append(upn)
        cls.__BPM_DOWNSTREAM = downnames
        cls.__BPM_UPSTREAM = upnames
        return cls.__BPM_NAMES

    @_classproperty
    def BPM_NICKNAMES(cls):
        """BPM nicknames."""
        cls.BPM_NAMES
        return cls.__BPM_NICKNAMES

    @_classproperty
    def BPM_POS(cls):
        """BPM positions."""
        cls.BPM_NAMES
        return cls.__BPM_POS

    @_classproperty
    def BPM_DOWNSTREAM(cls):
        """BPM downstream."""
        cls.BPM_NAMES
        return cls.__BPM_DOWNSTREAM

    @_classproperty
    def BPM_UPSTREAM(cls):
        """BPM upstream."""
        cls.BPM_NAMES
        return cls.__BPM_UPSTREAM

    def __init__(self):
        """Init."""
        self._oper = {'mean': self._mean, 'diff': self._diff, 'min': min}

    @staticmethod
    def get_down_up_bpms(bpmname):
        """Return down and up BPM names for bpmname."""
        index = BaseOrbitIntlk.BPM_NAMES.index(bpmname)
        down = BaseOrbitIntlk.BPM_DOWNSTREAM[index]
        upn = BaseOrbitIntlk.BPM_UPSTREAM[index]
        return down, upn

    def calc_intlk_metric(self, posarray, operation='', metric=''):
        """Return interlock metric, position or angulation."""
        if not operation:
            if not metric:
                raise ValueError(
                    'either the operation or the metric is required'
                )
            operation = 'mean' if 'pos' in metric.lower() else 'diff'

        data_values = list()
        for bpm in BaseOrbitIntlk.BPM_NAMES:
            down, upn = BaseOrbitIntlk.get_down_up_bpms(bpm)
            dval = posarray[BaseOrbitIntlk.BPM_NAMES.index(down)]
            uval = posarray[BaseOrbitIntlk.BPM_NAMES.index(upn)]
            func = self._oper[operation]
            val = func(dval, uval)
            data_values.append(val)
        return _np.array(data_values)

    @staticmethod
    def _mean(var1, var2):
        return (var1 + var2) / 2

    @staticmethod
    def _diff(var1, var2):
        return var1 - var2


class BPMOrbitIntlk(BaseOrbitIntlk, _Device):
    """This device group the orbit interlock PVs from one BPM."""

    PROPERTIES_DEFAULT = (
        # ==============================================================
        # General
        # **************************************************************
        # General interlock enable:
        'IntlkEn-Sel',
        'IntlkEn-Sts',
        # General interlock reset:
        'IntlkClr-Cmd',
        # Minimum sum threshold enable:
        # Enable orbit interlock only when sum is higher than IntlkLmtMinSum-RB
        'IntlkMinSumEn-Sel',
        'IntlkMinSumEn-Sts',
        # Minimum sum threshold (sum counts in FAcq rate):
        'IntlkLmtMinSum-SP',
        'IntlkLmtMinSum-RB',
        # Instantaneous interlock, difficult to be checked in the current
        # gateware implementation
        'Intlk-Mon',
        # Latch interlock, clean only when respective "Clr" PV is triggered
        'IntlkLtc-Mon',
        # ===============================================================
        # Position Interlock
        # ***************************************************************
        # Condition for position interlock:
        #   thres_min > (pos BPM downstream + pos BPM upstream)/2 or
        #   thres_max < (pos BPM downstream + pos BPM upstream)/2
        # BPMs are grouped 2 by 2 following the order seen by the beam:
        # - M1/M2
        # - C1-1/C1-2
        # - C2/C3-1
        # - C3-2/C4
        # upstream BPM is always the "first" BPM of the pairs above and
        # downstream BPM is always the "second" BPM of the pair.
        # ***************************************************************
        # Position interlock enable:
        'IntlkPosEn-Sel',
        'IntlkPosEn-Sts',
        # Position interlock clear:
        'IntlkPosClr-Cmd',
        # Thresholds (nm, FAcq rate):
        'IntlkLmtPosMaxX-SP',
        'IntlkLmtPosMaxX-RB',
        'IntlkLmtPosMinX-SP',
        'IntlkLmtPosMinX-RB',
        'IntlkLmtPosMaxY-SP',
        'IntlkLmtPosMaxY-RB',
        'IntlkLmtPosMinY-SP',
        'IntlkLmtPosMinY-RB',
        # All interlocks are masked by the "Enable" state
        # Instantaneous Status:
        'IntlkPosLowerX-Mon',
        'IntlkPosUpperX-Mon',  # X
        'IntlkPosLowerY-Mon',
        'IntlkPosUpperY-Mon',  # Y
        # Latch Status, clean only when respective "Clr" PV is triggered
        'IntlkPosLowerLtcX-Mon',
        'IntlkPosUpperLtcX-Mon',
        'IntlkPosLowerLtcY-Mon',
        'IntlkPosUpperLtcY-Mon',
        # Position measure from orbit interlock core
        'IntlkPosX-Mon',
        'IntlkPosY-Mon',
        # =============================================================
        # Angle Interlock
        # *************************************************************
        # Condition for angle interlock:
        #   thres_min > (pos BPM downstream - pos BPM upstream) or
        #   thres_max < (pos BPM downstream - pos BPM upstream)
        # BPMs are grouped 2 by 2 following the order seen by the beam:
        # - M1/M2
        # - C1-1/C1-2
        # - C2/C3-1
        # - C3-2/C4
        # upstream BPM is always the "first" BPM of the pairs above and
        # downstream BPM is always the "second" BPM of the pair.
        # ************************************************************
        # Angle interlock enable:
        'IntlkAngEn-Sel',
        'IntlkAngEn-Sts',
        # Angle interlock clear:
        'IntlkAngClr-Cmd',
        # Thresholds (rad.nm, FAcq rate).
        # Thresholds must be calculated as angle (rad)
        # between the 2 adjacent BPMs * distance (nm) between them:
        'IntlkLmtAngMaxX-SP',
        'IntlkLmtAngMaxX-RB',
        'IntlkLmtAngMinX-SP',
        'IntlkLmtAngMinX-RB',
        'IntlkLmtAngMaxY-SP',
        'IntlkLmtAngMaxY-RB',
        'IntlkLmtAngMinY-SP',
        'IntlkLmtAngMinY-RB',
        # All interlocks are masked by the "Enable" state
        # Intantaneous Status
        'IntlkAngLowerX-Mon',
        'IntlkAngUpperX-Mon',  # X
        'IntlkAngLowerY-Mon',
        'IntlkAngUpperY-Mon',  # Y
        # Latch Status, clean only when respective "Clr" PV is triggered
        'IntlkAngLowerLtcX-Mon',
        'IntlkAngUpperLtcX-Mon',
        'IntlkAngLowerLtcY-Mon',
        'IntlkAngUpperLtcY-Mon',
        # Angle measure from orbit interlock core
        'IntlkAngX-Mon',
        'IntlkAngY-Mon',
        # ============================================================
    )

    def __init__(self, devname, props2init='all'):
        """Init."""
        # call base class constructor
        BaseOrbitIntlk.__init__(self)
        if devname not in self.BPM_NAMES:
            raise ValueError(devname + ' is no a valid BPM name.')
        _Device.__init__(self, devname, props2init=props2init)

    @property
    def intlkposx(self):
        """Orbit interlock core Position X."""
        return self['IntlkPosX-Mon'] * self.CONV_NM2UM

    @property
    def intlkposy(self):
        """Orbit interlock core Position Y."""
        return self['IntlkPosY-Mon'] * self.CONV_NM2UM

    @property
    def intlkangx(self):
        """Orbit interlock core Angle X."""
        return self['IntlkAngX-Mon'] * self.CONV_NM2UM

    @property
    def intlkangy(self):
        """Orbit interlock core Angle Y."""
        return self['IntlkAngY-Mon'] * self.CONV_NM2UM

    @property
    def pair_down_up_bpms(self):
        """Return pair of downstream and upstream BPM names."""
        return self.get_down_up_bpms(self.devname)

    # --- general interlock ---

    @property
    def gen_enable(self):
        """General interlock enable."""
        return self['IntlkEn-Sts']

    @gen_enable.setter
    def gen_enable(self, value):
        self['IntlkEn-Sel'] = int(value)

    def cmd_reset_gen(self):
        """General interlock clear."""
        self['IntlkClr-Cmd'] = 1
        return True

    @property
    def gen_inst(self):
        """Instantaneous general interlock."""
        return self['Intlk-Mon']

    @property
    def gen_latch(self):
        """Latch general interlock."""
        return self['IntlkLtc-Mon']

    # --- minimum sum threshold ---

    @property
    def minsum_enable(self):
        """Minimum sum threshold enable.

        If enabled, generate orbit interlock only when sum exceeds
        value in 'minsum_thres' property.
        """
        return self['IntlkMinSumEn-Sts']

    @minsum_enable.setter
    def minsum_enable(self, value):
        self['IntlkMinSumEn-Sel'] = int(value)

    @property
    def minsum_thres(self):
        """Minimum sum threshold [sum count, FAcq rate]."""
        return self['IntlkLmtMinSum-RB']

    @minsum_thres.setter
    def minsum_thres(self, value):
        self['IntlkLmtMinSum-SP'] = int(value)

    # --- position interlock ---

    @property
    def pos_enable(self):
        """Position interlock enable."""
        return self['IntlkPosEn-Sts']

    @pos_enable.setter
    def pos_enable(self, value):
        self['IntlkPosEn-Sel'] = int(value)

    def cmd_reset_pos(self):
        """Position interlock clear."""
        self['IntlkPosClr-Cmd'] = 1
        return True

    @property
    def pos_x_min_thres(self):
        """Minimum X position threshold."""
        return self['IntlkLmtPosMinX-RB']

    @pos_x_min_thres.setter
    def pos_x_min_thres(self, value):
        self['IntlkLmtPosMinX-SP'] = value

    @property
    def pos_x_max_thres(self):
        """Maximum X position threshold."""
        return self['IntlkLmtPosMaxX-RB']

    @pos_x_max_thres.setter
    def pos_x_max_thres(self, value):
        self['IntlkLmtPosMaxX-SP'] = value

    @property
    def pos_y_min_thres(self):
        """Minimum Y position threshold."""
        return self['IntlkLmtPosMinY-RB']

    @pos_y_min_thres.setter
    def pos_y_min_thres(self, value):
        self['IntlkLmtPosMinY-SP'] = value

    @property
    def pos_y_max_thres(self):
        """Maximum Y position threshold."""
        return self['IntlkLmtPosMaxY-RB']

    @pos_y_max_thres.setter
    def pos_y_max_thres(self, value):
        self['IntlkLmtPosMaxY-SP'] = value

    @property
    def pos_inst_lower_x(self):
        """Instantaneous position interlock set when X
        minimum threshold is exceeded.
        """
        return self['IntlkPosLowerX-Mon']

    @property
    def pos_inst_upper_x(self):
        """Instantaneous position interlock set when X
        maximum threshold is exceeded.
        """
        return self['IntlkPosUpperX-Mon']

    @property
    def pos_inst_lower_y(self):
        """Instantaneous position interlock set when Y
        minimum threshold is exceeded.
        """
        return self['IntlkPosLowerY-Mon']

    @property
    def pos_inst_upper_y(self):
        """Instantaneous position interlock set when Y
        maximum threshold is exceeded.
        """
        return self['IntlkPosUpperY-Mon']

    @property
    def pos_latch_lower_x(self):
        """Latch position interlock set when X
        minimum threshold is exceeded.
        """
        return self['IntlkPosLowerLtcX-Mon']

    @property
    def pos_latch_upper_x(self):
        """Latch position interlock set when X
        maximum threshold is exceeded.
        """
        return self['IntlkPosUpperLtcX-Mon']

    @property
    def pos_latch_lower_y(self):
        """Latch position interlock set when Y
        minimum threshold is exceeded.
        """
        return self['IntlkPosLowerLtcY-Mon']

    @property
    def pos_latch_upper_y(self):
        """Latch position interlock set when Y
        maximum threshold is exceeded.
        """
        return self['IntlkPosUpperLtcY-Mon']

    #  --- angulation interlock ---

    @property
    def ang_enable(self):
        """Angulation interlock enable."""
        return self['IntlkAngEn-Sts']

    @ang_enable.setter
    def ang_enable(self, value):
        self['IntlkAngEn-Sel'] = int(value)

    def cmd_reset_ang(self):
        """Angulation interlock clear."""
        self['IntlkAngClr-Cmd'] = 1
        return True

    @property
    def ang_x_min_thres(self):
        """Minimum X angulation threshold."""
        return self['IntlkLmtAngMinX-RB']

    @ang_x_min_thres.setter
    def ang_x_min_thres(self, value):
        self['IntlkLmtAngMinX-SP'] = value

    @property
    def ang_x_max_thres(self):
        """Maximum X angulation threshold."""
        return self['IntlkLmtAngMaxX-RB']

    @ang_x_max_thres.setter
    def ang_x_max_thres(self, value):
        self['IntlkLmtAngMaxX-SP'] = value

    @property
    def ang_y_min_thres(self):
        """Minimum Y angulation threshold."""
        return self['IntlkLmtAngMinY-RB']

    @ang_y_min_thres.setter
    def ang_y_min_thres(self, value):
        self['IntlkLmtAngMinY-SP'] = value

    @property
    def ang_y_max_thres(self):
        """Maximum Y angulation threshold."""
        return self['IntlkLmtAngMaxY-RB']

    @ang_y_max_thres.setter
    def ang_y_max_thres(self, value):
        self['IntlkLmtAngMaxY-SP'] = value

    @property
    def ang_inst_lower_x(self):
        """Instantaneous angulation interlock set when X
        minimum threshold is exceeded.
        """
        return self['IntlkAngLowerX-Mon']

    @property
    def ang_inst_upper_x(self):
        """Instantaneous angulation interlock set when X
        maximum threshold is exceeded.
        """
        return self['IntlkAngUpperX-Mon']

    @property
    def ang_inst_lower_y(self):
        """Instantaneous angulation interlock set when Y
        minimum threshold is exceeded.
        """
        return self['IntlkAngLowerY-Mon']

    @property
    def ang_inst_upper_y(self):
        """Instantaneous angulation interlock set when Y
        maximum threshold is exceeded.
        """
        return self['IntlkAngUpperY-Mon']

    @property
    def ang_latch_lower_x(self):
        """Latch angulation interlock set when X
        minimum threshold is exceeded.
        """
        return self['IntlkAngLowerLtcX-Mon']

    @property
    def ang_latch_upper_x(self):
        """Latch angulation interlock set when X
        maximum threshold is exceeded.
        """
        return self['IntlkAngUpperLtcX-Mon']

    @property
    def ang_latch_lower_y(self):
        """Latch angulation interlock set when Y
        minimum threshold is exceeded.
        """
        return self['IntlkAngLowerLtcY-Mon']

    @property
    def ang_latch_upper_y(self):
        """Latch angulation interlock set when Y
        maximum threshold is exceeded.
        """
        return self['IntlkAngUpperLtcY-Mon']


class OrbitInterlock(BaseOrbitIntlk, _DeviceSet):
    """Orbit Interlock device."""

    TIMEOUT = 10

    class DEVICES:
        """."""

        SI = 'SI-Fam:DI-BPM'
        ALL = (SI,)

    def __init__(self, devname=None, props2init='all'):
        """Init."""
        if devname is None:
            devname = self.DEVICES.SI
        if devname not in self.DEVICES.ALL:
            raise ValueError('Wrong value for devname')
        BaseOrbitIntlk.__init__(self)
        devs = [
            BPMOrbitIntlk(dev, props2init=props2init) for dev in self.BPM_NAMES
        ]
        _DeviceSet.__init__(self, devs, devname=devname)

    # --- general interlock ---

    @property
    def gen_enable(self):
        """General interlock enable.

        Returns:
            enbl (numpy.ndarray, 160):
                enable state for each BPM.
        """
        return _np.array([b.gen_enable for b in self._devices])

    def set_gen_enable(self, value, timeout=TIMEOUT, return_prob=False):
        """Set enable state for BPM general interlock."""
        self.set_devices_propty('IntlkEn-Sel', value)
        return self.wait_devices_propty(
            'IntlkEn-Sts', value, timeout=timeout, return_prob=return_prob
        )

    def cmd_gen_enable(self, timeout=TIMEOUT, return_prob=False):
        """Enable all BPM general interlock."""
        for dev in self.devices:
            dev.gen_enable = 1
        return self.wait_devices_propty(
            'IntlkEn-Sts', 1, timeout=timeout, return_prob=return_prob
        )

    def cmd_gen_disable(self, timeout=TIMEOUT, return_prob=False):
        """Disable all BPM general interlock."""
        for dev in self.devices:
            dev.gen_enable = 0
        return self.wait_devices_propty(
            'IntlkEn-Sts', 0, timeout=timeout, return_prob=return_prob
        )

    def cmd_reset_gen(self):
        """Reset all BPM general interlock."""
        for dev in self.devices:
            dev.cmd_reset_gen()
        return True

    @property
    def gen_inst(self):
        """General instantaneous interlocks.

        Returns:
            gen_inst (numpy.ndarray, 160):
                general instantaneous interlock for each BPM.
        """
        return _np.array([b.gen_inst for b in self._devices])

    @property
    def gen_latch(self):
        """General latch interlocks.

        Returns:
            gen_latch (numpy.ndarray, 160):
                general latch interlock for each BPM.
        """
        return _np.array([b.gen_latch for b in self._devices])

    # --- minimum sum threshold ---

    @property
    def minsum_enable(self):
        """Minimum sum threshold enable.

        Returns:
            enbl (numpy.ndarray, 160):
                enable state for each BPM.
        """
        return _np.array([b.minsum_enable for b in self._devices])

    def set_minsum_enable(self, value, timeout=TIMEOUT, return_prob=False):
        """Set enable state for BPM minimum sum interlock."""
        self.set_devices_propty('IntlkMinSumEn-Sel', value)
        return self.wait_devices_propty(
            'IntlkMinSumEn-Sts',
            value,
            timeout=timeout,
            return_prob=return_prob,
        )

    def cmd_minsum_enable(self, timeout=TIMEOUT, return_prob=False):
        """Enable all BPM minimum sum threshold."""
        for dev in self.devices:
            dev.minsum_enable = 1
        return self.wait_devices_propty(
            'IntlkMinSumEn-Sts', 1, timeout=timeout, return_prob=return_prob
        )

    def cmd_minsum_disable(self, timeout=TIMEOUT, return_prob=False):
        """Disable all BPM minimum sum threshold."""
        for dev in self.devices:
            dev.minsum_enable = 0
        return self.wait_devices_propty(
            'IntlkMinSumEn-Sts', 0, timeout=timeout, return_prob=return_prob
        )

    @property
    def minsum_thres(self):
        """Minimum sum thresholds.

        Returns:
            thres (numpy.ndarray, 160): min.sum threshold for each BPM.
        """
        return _np.array([b.minsum_thres for b in self._devices])

    def set_minsum_thres(self, value, timeout=TIMEOUT, return_prob=False):
        """Set minimum sum thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum sum.
            timeout (float, optional): timeout to wait. Defaults to TIMEOUT.

        Returns:
            bool: True if set was successful.

        """
        value = self._handle_thres_input(value)
        self.set_devices_propty('IntlkLmtMinSum-SP', value)
        return self.wait_devices_propty(
            'IntlkLmtMinSum-RB',
            value,
            timeout=timeout,
            return_prob=return_prob,
        )

    # --- position interlock ---

    @property
    def pos_enable(self):
        """Position interlock enable.

        Returns:
            enbl (numpy.ndarray, 160):
                enable state for each BPM.
        """
        return _np.array([b.pos_enable for b in self._devices])

    def set_pos_enable(self, value, timeout=TIMEOUT, return_prob=False):
        """Set enable state for BPM position interlock."""
        self.set_devices_propty('IntlkPosEn-Sel', value)
        return self.wait_devices_propty(
            'IntlkPosEn-Sts', value, timeout=timeout, return_prob=return_prob
        )

    def cmd_pos_enable(self, timeout=TIMEOUT, return_prob=False):
        """Enable all BPM position interlock."""
        for dev in self.devices:
            dev.pos_enable = 1
        return self.wait_devices_propty(
            'IntlkPosEn-Sts', 1, timeout=timeout, return_prob=return_prob
        )

    def cmd_pos_disable(self, timeout=TIMEOUT, return_prob=False):
        """Disable all BPM position interlock."""
        for dev in self.devices:
            dev.pos_enable = 0
        return self.wait_devices_propty(
            'IntlkPosEn-Sts', 0, timeout=timeout, return_prob=return_prob
        )

    def cmd_reset_pos(self):
        """Reset all BPM position interlock."""
        for dev in self.devices:
            dev.cmd_reset_pos()
        return True

    @property
    def pos_x_min_thres(self):
        """Minimum x position thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. x position threshold for each BPM.
        """
        return _np.array([b.pos_x_min_thres for b in self._devices])

    def set_pos_x_min_thres(self, value, timeout=TIMEOUT, return_prob=False):
        """Set minimum x position thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum sum.
            timeout (float, optional): timeout to wait. Defaults to TIMEOUT.

        Returns:
            bool: True if set was successful.

        """
        value = self._handle_thres_input(value)
        self.set_devices_propty('IntlkLmtPosMinX-SP', value)
        return self.wait_devices_propty(
            'IntlkLmtPosMinX-RB',
            value,
            timeout=timeout,
            return_prob=return_prob,
        )

    @property
    def pos_x_max_thres(self):
        """Maximum x position thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. x position threshold for each BPM.
        """
        return _np.array([b.pos_x_max_thres for b in self._devices])

    def set_pos_x_max_thres(self, value, timeout=TIMEOUT, return_prob=False):
        """Set maximum x position thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum sum.
            timeout (float, optional): timeout to wait. Defaults to TIMEOUT.

        Returns:
            bool: True if set was successful.

        """
        value = self._handle_thres_input(value)
        self.set_devices_propty('IntlkLmtPosMaxX-SP', value)
        return self.wait_devices_propty(
            'IntlkLmtPosMaxX-RB',
            value,
            timeout=timeout,
            return_prob=return_prob,
        )

    @property
    def pos_y_min_thres(self):
        """Minimum y position thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. y position threshold for each BPM.
        """
        return _np.array([b.pos_y_min_thres for b in self._devices])

    def set_pos_y_min_thres(self, value, timeout=TIMEOUT, return_prob=False):
        """Set minimum y position thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum sum.
            timeout (float, optional): timeout to wait. Defaults to TIMEOUT.

        Returns:
            bool: True if set was successful.

        """
        value = self._handle_thres_input(value)
        self.set_devices_propty('IntlkLmtPosMinY-SP', value)
        return self.wait_devices_propty(
            'IntlkLmtPosMinY-RB',
            value,
            timeout=timeout,
            return_prob=return_prob,
        )

    @property
    def pos_y_max_thres(self):
        """Maximum y position thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. y position threshold for each BPM.
        """
        return _np.array([b.pos_y_max_thres for b in self._devices])

    def set_pos_y_max_thres(self, value, timeout=TIMEOUT, return_prob=False):
        """Set maximum y position thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum sum.
            timeout (float, optional): timeout to wait. Defaults to TIMEOUT.

        Returns:
            bool: True if set was successful.

        """
        value = self._handle_thres_input(value)
        self.set_devices_propty('IntlkLmtPosMaxY-SP', value)
        return self.wait_devices_propty(
            'IntlkLmtPosMaxY-RB',
            value,
            timeout=timeout,
            return_prob=return_prob,
        )

    @property
    def pos_inst_lower_x(self):
        """Instantaneous position interlock set when X
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_inst_lower_x for b in self._devices])

    @property
    def pos_inst_upper_x(self):
        """Instantaneous position interlock set when X
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_inst_upper_x for b in self._devices])

    @property
    def pos_inst_lower_y(self):
        """Instantaneous position interlock set when Y
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_inst_lower_y for b in self._devices])

    @property
    def pos_inst_upper_y(self):
        """Instantaneous position interlock set when Y
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_inst_upper_y for b in self._devices])

    @property
    def pos_latch_lower_x(self):
        """Latch position interlock set when X
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_latch_lower_x for b in self._devices])

    @property
    def pos_latch_upper_x(self):
        """Latch position interlock set when X
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_latch_upper_x for b in self._devices])

    @property
    def pos_latch_lower_y(self):
        """Latch position interlock set when Y
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_latch_lower_y for b in self._devices])

    @property
    def pos_latch_upper_y(self):
        """Latch position interlock set when Y
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.pos_latch_upper_y for b in self._devices])

    #  --- angulation interlock ---

    @property
    def ang_enable(self):
        """Angle interlock enable.

        Returns:
            enbl (numpy.ndarray, 160):
                enable state for each BPM.
        """
        return _np.array([b.ang_enable for b in self._devices])

    def set_ang_enable(self, value, timeout=TIMEOUT, return_prob=False):
        """Set enable state for BPM angulation interlock."""
        self.set_devices_propty('IntlkAngEn-Sel', value)
        return self.wait_devices_propty(
            'IntlkAngEn-Sts', value, timeout=timeout, return_prob=return_prob
        )

    def cmd_ang_enable(self, timeout=TIMEOUT, return_prob=False):
        """Enable all BPM angulation interlock."""
        for dev in self.devices:
            dev.ang_enable = 1
        return self.wait_devices_propty(
            'IntlkAngEn-Sts', 1, timeout=timeout, return_prob=return_prob
        )

    def cmd_ang_disable(self, timeout=TIMEOUT, return_prob=False):
        """Disable all BPM angulation interlock."""
        for dev in self.devices:
            dev.ang_enable = 0
        return self.wait_devices_propty(
            'IntlkAngEn-Sts', 0, timeout=timeout, return_prob=return_prob
        )

    def cmd_reset_ang(self):
        """Reset all BPM angulation interlock."""
        for dev in self.devices:
            dev.cmd_reset_ang()
        return True

    @property
    def ang_x_min_thres(self):
        """Minimum x angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. x angulation threshold for each BPM.
        """
        return _np.array([b.ang_x_min_thres for b in self._devices])

    def set_ang_x_min_thres(self, value, timeout=TIMEOUT, return_prob=False):
        """Set minimum x angle thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum sum.
            timeout (float, optional): timeout to wait. Defaults to TIMEOUT.

        Returns:
            bool: True if set was successful.

        """
        value = self._handle_thres_input(value)
        self.set_devices_propty('IntlkLmtAngMinX-SP', value)
        return self.wait_devices_propty(
            'IntlkLmtAngMinX-RB',
            value,
            timeout=timeout,
            return_prob=return_prob,
        )

    @property
    def ang_x_max_thres(self):
        """Maximum x angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. x angulation threshold for each BPM.
        """
        return _np.array([b.ang_x_max_thres for b in self._devices])

    def set_ang_x_max_thres(self, value, timeout=TIMEOUT, return_prob=False):
        """Set maximum x angle thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum sum.
            timeout (float, optional): timeout to wait. Defaults to TIMEOUT.

        Returns:
            bool: True if set was successful.

        """
        value = self._handle_thres_input(value)
        self.set_devices_propty('IntlkLmtAngMaxX-SP', value)
        return self.wait_devices_propty(
            'IntlkLmtAngMaxX-RB',
            value,
            timeout=timeout,
            return_prob=return_prob,
        )

    @property
    def ang_y_min_thres(self):
        """Minimum y angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. y angulation threshold for each BPM.
        """
        return _np.array([b.ang_y_min_thres for b in self._devices])

    def set_ang_y_min_thres(self, value, timeout=TIMEOUT, return_prob=False):
        """Set minimum y angle thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum sum.
            timeout (float, optional): timeout to wait. Defaults to TIMEOUT.

        Returns:
            bool: True if set was successful.

        """
        value = self._handle_thres_input(value)
        self.set_devices_propty('IntlkLmtAngMinY-SP', value)
        return self.wait_devices_propty(
            'IntlkLmtAngMinY-RB',
            value,
            timeout=timeout,
            return_prob=return_prob,
        )

    @property
    def ang_y_max_thres(self):
        """Maximum y angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. y angulation threshold for each BPM.
        """
        return _np.array([b.ang_y_max_thres for b in self._devices])

    def set_ang_y_max_thres(self, value, timeout=TIMEOUT, return_prob=False):
        """Set maximum y angle thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum sum.
            timeout (float, optional): timeout to wait. Defaults to TIMEOUT.

        Returns:
            bool: True if set was successful.

        """
        value = self._handle_thres_input(value)
        self.set_devices_propty('IntlkLmtAngMaxY-SP', value)
        return self.wait_devices_propty(
            'IntlkLmtAngMaxY-RB',
            value,
            timeout=timeout,
            return_prob=return_prob,
        )

    @property
    def ang_inst_lower_x(self):
        """Instantaneous angulation interlock set when X
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_lower_x for b in self._devices])

    @property
    def ang_inst_upper_x(self):
        """Instantaneous angulation interlock set when X
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_upper_x for b in self._devices])

    @property
    def ang_inst_lower_y(self):
        """Instantaneous angulation interlock set when Y
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_lower_y for b in self._devices])

    @property
    def ang_inst_upper_y(self):
        """Instantaneous angulation interlock set when Y
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_upper_y for b in self._devices])

    @property
    def ang_latch_lower_x(self):
        """Latch angulation interlock set when X
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_lower_x for b in self._devices])

    @property
    def ang_latch_upper_x(self):
        """Latch angulation interlock set when X
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_upper_x for b in self._devices])

    @property
    def ang_latch_lower_y(self):
        """Latch angulation interlock set when Y
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_lower_y for b in self._devices])

    @property
    def ang_latch_upper_y(self):
        """Latch angulation interlock set when Y
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_upper_y for b in self._devices])

    @property
    def position(self):
        """Position vectors.

        Position at each BPM is defined as:
            (pos BPM downstream + pos BPM upstream)/2

        Returns:
            posx (numpy.ndarray, 160): Horizontal Position.
            posy (numpy.ndarray, 160): Vertical Position.
        """
        posx = _np.array([b.intlkposx for b in self._devices])
        posy = _np.array([b.intlkposy for b in self._devices])
        return posx, posy

    @property
    def angle(self):
        """Angle vectors.

        Angle at each BPM is defined as:
            (posição BPM downstream - posição BPM upstream)

        Returns:
            angx (numpy.ndarray, 160): Horizontal Angle.
            angy (numpy.ndarray, 160): Vertical Angle.
        """
        angx = _np.array([b.intlkangx for b in self._devices])
        angy = _np.array([b.intlkangy for b in self._devices])
        return angx, angy

    def _handle_thres_input(self, value):
        typs = (bool, int, float, _np.bool_, _np.int64, _np.float64)
        if isinstance(value, typs):
            return [value] * len(self.BPM_NAMES)
        if len(value) != len(self.BPM_NAMES):
            raise ValueError(
                f'Expected length is {len(self.BPM_NAMES)}, not {len(value)}'
            )
        return value


class HLOrbitInterlock(BaseOrbitIntlk, _Device):
    """."""

    TIMEOUT = 10

    class DEVICES:
        """."""

        SI = 'SI-Glob:AP-OrbIntlk'
        ALL = (SI,)

    PROPERTIES_DEFAULT = (
        'Enable-Sel',
        'Enable-Sts',
        'Reset-Cmd',
        'BPMStatus-Mon',
        'TimingStatus-Mon',
        'LLRFStatus-Mon',
        'PosXMaxLim-SP',
        'PosXMaxLim-RB',
        'PosXMinLim-SP',
        'PosXMinLim-RB',
        'AngXMaxLim-SP',
        'AngXMaxLim-RB',
        'AngXMinLim-SP',
        'AngXMinLim-RB',
        'PosYMaxLim-SP',
        'PosYMaxLim-RB',
        'PosYMinLim-SP',
        'PosYMinLim-RB',
        'AngYMaxLim-SP',
        'AngYMaxLim-RB',
        'AngYMinLim-SP',
        'AngYMinLim-RB',
        'MinSumLim-SP',
        'MinSumLim-RB',
        'PosEnblList-SP',
        'PosEnblList-RB',
        'AngEnblList-SP',
        'AngEnblList-RB',
        'MinSumEnblList-SP',
        'MinSumEnblList-RB',
    )

    def __init__(self, props2init='all'):
        """."""
        _Device.__init__(self, devname=self.DEVICES.SI, props2init=props2init)
        BaseOrbitIntlk.__init__(self)

    @property
    def enable(self):
        """General interlock enable.

        Returns:
            enbl (numpy.ndarray, 160): enable state for each BPM.
        """
        return self['Enable-Sts']

    def set_enable(self, value, timeout=TIMEOUT):
        """Set enable state for BPM general interlock."""
        self['Enable-Sel'] = bool(value)
        return self.wait('Enable-Sts', bool(value), timeout=timeout)

    def cmd_enable(self, timeout=TIMEOUT):
        """Enable all BPM general interlock."""
        return self.set_enable(True, timeout=timeout)

    def cmd_disable(self, timeout=TIMEOUT):
        """Disable all BPM general interlock."""
        return self.set_enable(False, timeout=timeout)

    def cmd_reset(self, timeout=TIMEOUT):
        """Enable all BPM general interlock."""
        val = self['Reset-Cmd']
        self['Reset-Cmd'] = True
        return self.wait('Reset-Cmd', val + 1, timeout=timeout)

    @property
    def pos_x_min_thres(self):
        """Minimum x position thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. x position threshold for each BPM.
        """
        return self['PosXMinLim-RB']

    @pos_x_min_thres.setter
    def pos_x_min_thres(self, value):
        """Set minimum x position thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum x position.
        """
        self['PosXMinLim-SP'] = self._handle_thres_input(value)

    @property
    def pos_x_max_thres(self):
        """Maximum x position thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. x position threshold for each BPM.
        """
        return self['PosXMaxLim-RB']

    @pos_x_max_thres.setter
    def pos_x_max_thres(self, value):
        """Set maximum x position thresholds.

        Args:
            value (numpy.ndarray|float): Values for maximum x position.
        """
        self['PosXMaxLim-SP'] = self._handle_thres_input(value)

    @property
    def pos_y_min_thres(self):
        """Minimum y position thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. y position threshold for each BPM.
        """
        return self['PosYMinLim-RB']

    @pos_y_min_thres.setter
    def pos_y_min_thres(self, value):
        """Set minimum y position thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum y position.
        """
        self['PosYMinLim-SP'] = self._handle_thres_input(value)

    @property
    def pos_y_max_thres(self):
        """Maximum y position thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. y position threshold for each BPM.
        """
        return self['PosYMaxLim-RB']

    @pos_y_max_thres.setter
    def pos_y_max_thres(self, value):
        """Set maximum y position thresholds.

        Args:
            value (numpy.ndarray|float): Values for maximum y position.
        """
        self['PosYMaxLim-SP'] = self._handle_thres_input(value)

    @property
    def ang_x_min_thres(self):
        """Minimum x angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. x angulation threshold for each BPM.
        """
        return self['AngXMinLim-RB']

    @ang_x_min_thres.setter
    def ang_x_min_thres(self, value):
        """Set minimum x angle thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum x angle.
        """
        self['AngXMinLim-SP'] = self._handle_thres_input(value)

    @property
    def ang_x_max_thres(self):
        """Maximum x angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. x angulation threshold for each BPM.
        """
        return self['AngXMaxLim-RB']

    @ang_x_max_thres.setter
    def ang_x_max_thres(self, value):
        """Set maximum x angle thresholds.

        Args:
            value (numpy.ndarray|float): Values for maximum x angle.
        """
        self['AngXMaxLim-SP'] = self._handle_thres_input(value)

    @property
    def ang_y_min_thres(self):
        """Minimum y angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. y angulation threshold for each BPM.
        """
        return self['AngYMinLim-RB']

    @ang_y_min_thres.setter
    def ang_y_min_thres(self, value):
        """Set minimum y angle thresholds.

        Args:
            value (numpy.ndarray|float): Values for minimum y angle.
        """
        self['AngYMinLim-SP'] = self._handle_thres_input(value)

    @property
    def ang_y_max_thres(self):
        """Maximum y angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. y angulation threshold for each BPM.
        """
        return self['AngYMaxLim-RB']

    @ang_y_max_thres.setter
    def ang_y_max_thres(self, value):
        """Set maximum y angle thresholds.

        Args:
            value (numpy.ndarray|float): Values for maximum y angle.
        """
        self['AngYMaxLim-SP'] = self._handle_thres_input(value)

    @property
    def minsum_thres(self):
        """Min sum thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min sum threshold for each BPM.
        """
        return self['MinSumLim-RB']

    @minsum_thres.setter
    def minsum_thres(self, value):
        """Set min sum thresholds.

        Args:
            value (numpy.ndarray|float): Values for min sum.
        """
        self['MinSumLim-SP'] = self._handle_thres_input(value)

    @property
    def pos_enbl_list(self):
        """Position enable list.

        Returns:
            enbl_list (numpy.ndarray, 160): position enable list for each BPM.
        """
        return self['PosEnblList-RB']

    @pos_enbl_list.setter
    def pos_enbl_list(self, value):
        """Set position enable list.

        Args:
            value (numpy.ndarray|bool): Values for enable list.
        """
        self['PosEnblList-SP'] = self._handle_thres_input(value)

    @property
    def ang_enbl_list(self):
        """Angle enable list.

        Returns:
            enbl_list (numpy.ndarray, 160): angle enable list for each BPM.
        """
        return self['AngEnblList-RB']

    @ang_enbl_list.setter
    def ang_enbl_list(self, value):
        """Set angle enable list.

        Args:
            value (numpy.ndarray|bool): Values for enable list.
        """
        self['AngEnblList-SP'] = self._handle_thres_input(value)

    @property
    def minsum_enbl_list(self):
        """Min sum enable list.

        Returns:
            enbl_list (numpy.ndarray, 160): min sum enable list for each BPM.
        """
        return self['MinSumEnblList-RB']

    @minsum_enbl_list.setter
    def minsum_enbl_list(self, value):
        """Set min sum enable list.

        Args:
            value (numpy.ndarray|bool): Values for enable list.
        """
        self['MinSumEnblList-SP'] = self._handle_thres_input(value)

    def get_standard_interlock_limits_definitions(self):
        """Get default interlock limits from control-system-constants table."""
        return _OrbIntlkSearch.get_limits()

    def set_interlock_limits(self, refx=None, refy=None, limits=None):
        """Set interlock limits based on input config or default values.

        Args:
            refx (numpy.ndarray, optional): Horizontal reference orbit around
                which limits will be calculated. Defaults to None. If None,
                then the current 'bba_orb' will be used.
            refy (numpy.ndarray, optional): Vertical reference orbit around
                which limits will be calculated. Defaults to None.If None,
                then the current 'bba_orb' will be used.
            limits (dict, optional): Dictionary whose keys are the name of the
                subsection (ex. '01SA', '04BC') where to apply the interlock
                limits and values are dictionaries with keys:
                    'agx', 'agy', 'psx', 'psy'
                defining the values of the interlock limits around the
                reference orbit. The rest of the ring will have its limits set
                to 1.2 times the largest respective limit. Defaults to None.
                If None, then the standard definition from the
                control-system-constants table will be used.

        Raises:
            ValueError: In case inputs don't match the description above.
            TypeError: In case inputs don't match the description above.
            RunTimeError: In case interlock is enabled.
        """
        if self.enable:
            raise RuntimeError(
                'Cannot set limits, orbit interlock is enabled.'
            )
        if refx is None or refy is None:
            clt = ConfigDBClient(config_type='si_orbit')
            refo = clt.get_config_value('bba_orb')
            if refx is None:
                refx = _np.array(refo['x'])
            if refy is None:
                refy = _np.array(refo['y'])
        if limits is None:
            limits = self.get_standard_interlock_limits_definitions()

        if {len(refx), len(refy)} - {len(self.BPM_NAMES)}:
            raise ValueError('Length of reference orbit must match BPMs.')
        if not isinstance(limits, dict):
            raise TypeError('Limits must be a dictionary')

        entries = {'psx', 'psy', 'agx', 'agy'}
        for k, v in limits.items():
            if not isinstance(k, str):
                raise ValueError('Every key of limits must be a str.')
            if not isinstance(v, dict):
                raise TypeError('Every entry of limits must be dictionary.')
            if not v.keys() == entries:
                raise ValueError(
                    'Every entry of limits must have keys: '
                    + ', '.join(entries)
                )

        limits = limits.copy()

        flag = _np.zeros(refx.size, dtype=bool)
        zer = refx * 0.0
        agx, agy = zer.copy(), zer.copy()
        psx, psy = zer.copy(), zer.copy()
        mxp, myp, mxa, mya = 0.0, 0.0, 0.0, 0.0
        for ssec, idd in limits.items():
            # Position
            ids = idd.copy()
            [ids.pop(k) for k in ('agx', 'agy')]
            bpx, bpy = _si_calculate_bump(zer, zer, subsec=ssec, **ids)
            flg = bpx != 0.0
            psx[flg] = bpx.max()
            psy[flg] = bpy.max()
            mxp, myp = max(mxp, bpx.max()), max(myp, bpy.max())

            # Angle
            ids = idd.copy()
            [ids.pop(k) for k in ('psx', 'psy')]
            bpx, bpy = _si_calculate_bump(zer, zer, subsec=ssec, **ids)
            agx[flg] = bpx.max() * 2  # NOTE: not valid for BC, C1, C2. Needs
            agy[flg] = bpy.max() * 2  # generalization of BPM firmware.
            mxa, mya = max(mxa, bpx.max() * 2), max(mya, bpy.max() * 2)

            flag |= flg

        psx[~flag] = mxp * 1.2
        psy[~flag] = myp * 1.2
        agx[~flag] = mxa * 1.2
        agy[~flag] = mya * 1.2

        # set limits
        refx_pos = self.calc_intlk_metric(refx, metric='pos')
        refy_pos = self.calc_intlk_metric(refy, metric='pos')
        refx_ang = self.calc_intlk_metric(refx, metric='ang')
        refy_ang = self.calc_intlk_metric(refy, metric='ang')

        # needs convertion from [um] to [nm]:
        self.pos_x_max_thres = _np.round((refx_pos + psx) * 1e3)
        self.pos_x_min_thres = _np.round((refx_pos - psx) * 1e3)
        self.pos_y_max_thres = _np.round((refy_pos + psy) * 1e3)
        self.pos_y_min_thres = _np.round((refy_pos - psy) * 1e3)
        self.ang_x_max_thres = _np.round((refx_ang + agx) * 1e3)
        self.ang_x_min_thres = _np.round((refx_ang - agx) * 1e3)
        self.ang_y_max_thres = _np.round((refy_ang + agy) * 1e3)
        self.ang_y_min_thres = _np.round((refy_ang - agy) * 1e3)

        self.pos_enbl_list = flag
        self.ang_enbl_list = flag

    def _handle_thres_input(self, value):
        return OrbitInterlock._handle_thres_input(self, value)



"""Define properties of orbit interlock for the SIRIUS storage ring."""

import ast as _ast
from copy import deepcopy as _dcopy
from threading import Lock as _Lock

from siriuspy import clientweb as _web

_timeout = 1.0


class OrbIntlkSearch:
    """Contain properties of the triggers."""

    _limits = dict()
    _lock = _Lock()

    @classmethod
    def get_limits(cls):
        """Return a dictionary with high level events."""
        cls._init()
        return _dcopy(cls._limits)

    @classmethod
    def reset(cls):
        """Reset data."""
        cls._limits = dict()
        cls._init()

    @classmethod
    def _init(cls):
        """Initialize the Instance."""
        with cls._lock:
            if cls._limits:
                return
            if not _web.server_online():
                raise Exception('Could not connect with Consts Server!!')
            text = _web.orb_intlk_limits_read(timeout=_timeout)
            cls._limits = _ast.literal_eval(text)



import numpy as _np
from .csdev import HLFOFBConst as _Const
from ..devices import FamFOFBControllers as _FamFOFBCtrls, \
        FamFastCorrs as _FamFastCorrs


def get_default_psconfig_mat(option=None):
    """Set corrector accumulator filter.

    Args:
        option (str): either 'Switching' or 'Unit'.
            If None, we consider option = 'Unit'.

    Returns:
        mat (numpy.ndarray):
            PS Configuration matrix.

    """

    _const = _Const()
    _llfofb_dev = _FamFOFBCtrls()

    corrnames = _const.ch_names + _const.cv_names
    _corrs_dev = _FamFastCorrs(corrnames)

    num_biquads = _llfofb_dev.fofbacc_filter_num_biquads
    unit = _const.FILTER_UNIT
    sw2 = _const.FILTER_SW_2
    sw4 = _const.FILTER_SW_4
    unit_gain = _np.ones(_const.nr_chcv, dtype=float)

    if option == 'Unit' or option is None:
        _acc_filter_coeffs = num_biquads * unit

    elif option == 'Switching':
        _acc_filter_coeffs = sw2 + sw4 + (num_biquads - 2) * unit

    else:
        raise ValueError('Option must be Unit or Switching.')

    _corr_currloop_kp = _corrs_dev.currloop_kp
    _corr_currloop_ki = _corrs_dev.currloop_ki

    mat = _np.zeros(
        (_const.nr_chcv, _const.psconfig_nr_coeffs_columns +
         _const.PSCONFIG_COEFF_FIRST_COL),
        dtype=float
    )
    mat[:, _const.PSCONFIG_KP_COL] = _corr_currloop_kp
    mat[:, _const.PSCONFIG_KI_COL] = _corr_currloop_ki
    mat[:, _const.PSCONFIG_FILTER_GAIN_COL] = unit_gain
    mat[:, _const.PSCONFIG_COEFF_FIRST_COL:] = _acc_filter_coeffs

    return mat



"""InjSys Standby Handler devices."""

import logging as _log
import time as _time
from copy import deepcopy as _dcopy
from threading import Event as _Flag, Thread as _Thread

from ..callbacks import Callback as _Callback
from ..csdev import Const as _Const
from ..injctrl.csdev import Const as _InjConst
from ..pwrsupply.csdev import Const as _PSConst
from ..search import HLTimeSearch, PSSearch
from ..timesys.csdev import Const as _TIConst
from .device import Device as _Device, DeviceSet as _DeviceSet
from .lillrf import DevLILLRF
from .modltr import LIModltr
from .posang import PosAng
from .pwrsupply import PowerSupply, PowerSupplyPU
from .rf import ASLLRF
from .timing import HLTiming


class _BaseHandler(_DeviceSet):
    """Base standby mode handler for injection procedure."""

    def __init__(self, devices, hltiming=None):
        """."""
        self._hltiming = hltiming or HLTiming()
        devices = tuple(devices) + (self._hltiming,)
        super().__init__(devices)

    @property
    def hltiming(self):
        """."""
        return self._hltiming

    @hltiming.setter
    def hltiming(self, hltiming):
        if isinstance(hltiming, HLTiming):
            self._hltiming = hltiming

    @property
    def is_on(self):
        """Is on."""
        for dev, ppties in self._on_values.items():
            for ppty, val in ppties.items():
                if not dev[ppty] == val:
                    return False
        return True

    @property
    def on_values(self):
        """Property values equivalent to 'on' status."""
        vals = dict()
        for dev, ppties in self._on_values.items():
            for ppty, val in ppties.items():
                vals[dev.pv_object(ppty).pvname] = val
        return vals


class PUMagsStandbyHandler(_BaseHandler):
    """Pulsed magnets standby mode handler for injection procedure."""

    def __init__(self, hltiming=None):
        """Init."""
        self._punames = PSSearch.get_psnames({
            'dis': 'PU',
            'dev': '.*(Kckr|Sept)',
            'propty_name': '(?!:CCoil).*',
        })
        self._trignames = [dev.replace('PU', 'TI') for dev in self._punames]

        # pu devices
        self._pudevs = [PowerSupplyPU(pun) for pun in self._punames]

        # call base class constructor
        super().__init__(self._pudevs, hltiming=hltiming)

        # trigger devices
        self._trigdevs = [self._hltiming.triggers[t] for t in self._trignames]

        self._on_values = dict()
        for pudev in self._pudevs:
            if 'InjDpKckr' in pudev.devname:
                continue
            self._on_values[pudev] = {
                'Pulse-Sts': _PSConst.DsblEnbl.Enbl,
                'PwrState-Sts': _PSConst.DsblEnbl.Enbl,
            }
        for tdev in self._trigdevs:
            if 'InjDpKckr' in tdev.devname:
                continue
            self._on_values[tdev] = {'State-Sts': _TIConst.DsblEnbl.Enbl}

    @property
    def punames(self):
        """Power supply names."""
        return _dcopy(self._punames)

    @property
    def pudevices(self):
        """Power supply devices."""
        return self._pudevs

    @property
    def trignames(self):
        """Trigger names."""
        return _dcopy(self._trignames)

    @property
    def trigdevices(self):
        """Trigger devices."""
        return self._trigdevs

    def cmd_turn_off(self):
        """Turn off."""
        # set pulsed magnet pulse off
        self.set_devices_propty(
            'Pulse-Sel', _PSConst.DsblEnbl.Dsbl, devices=self._pudevs
        )

        # wait for pulsed magnet pulse to be off
        retval = self.wait_devices_propty(
            'Pulse-Sts',
            _PSConst.DsblEnbl.Dsbl,
            devices=self._pudevs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for pulsed magnet Pulse to be disabled '
                'timed out without success! Verify pulsed magnets!'
            )
            return [False, text, retval[1]]

        # set pulsed magnet power state off
        self.set_devices_propty(
            'PwrState-Sel', _PSConst.DsblEnbl.Dsbl, devices=self._pudevs
        )

        # wait for pulsed magnet power state to be off
        retval = self.wait_devices_propty(
            'PwrState-Sts',
            _PSConst.DsblEnbl.Dsbl,
            devices=self._pudevs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for pulsed magnet PwrState to be off '
                'timed out without success! Verify pulsed magnets!'
            )
            return [False, text, retval[1]]

        return True, '', []

    def cmd_turn_on(self):
        """Turn on."""
        devs = [dev for dev in self._pudevs if 'InjDpKckr' not in dev.devname]

        # set pulsed magnet power state on
        self.set_devices_propty(
            'PwrState-Sel', _PSConst.DsblEnbl.Enbl, devices=devs
        )

        # wait for pulsed magnet power state to be on
        retval = self.wait_devices_propty(
            'PwrState-Sts',
            _PSConst.DsblEnbl.Enbl,
            devices=devs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for pulsed magnet PwrState to be on '
                'timed out without success! Verify pulsed magnets!'
            )
            return [False, text, retval[1]]

        # wait a moment for the PU high voltage
        _time.sleep(1)

        # set pulsed magnet pulse on
        self.set_devices_propty(
            'Pulse-Sel', _PSConst.DsblEnbl.Enbl, devices=devs
        )

        # wait for pulsed magnet pulse to be on
        retval = self.wait_devices_propty(
            'Pulse-Sts',
            _PSConst.DsblEnbl.Enbl,
            devices=devs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for pulsed magnet Pulse to be enabled '
                'timed out without success! Verify pulsed magnets!'
            )
            return [False, text, retval[1]]

        return True, '', []


class BOPSRampStandbyHandler(_BaseHandler):
    """BO PS Ramp standy mode handler for injection procedure."""

    def __init__(self, hltiming=None):
        """Init."""
        self._psnames = PSSearch.get_psnames({'sec': 'BO', 'dis': 'PS'})
        self._trignames = HLTimeSearch.get_hl_triggers({
            'sec': 'BO',
            'dev': 'Mags',
        })

        # ps devices
        self._psdevs = [PowerSupply(psn) for psn in self._psnames]

        # call base class constructor
        super().__init__(self._psdevs, hltiming=hltiming)

        # trigger devices
        self._trigdevs = [self._hltiming.triggers[t] for t in self._trignames]

        self._on_values = dict()
        for psdev in self._psdevs:
            self._on_values[psdev] = {
                'OpMode-Sts': _PSConst.States.RmpWfm,
                'WfmUpdateAuto-Sts': _PSConst.DsblEnbl.Enbl,
            }
        for tdev in self._trigdevs:
            self._on_values[tdev] = {
                'State-Sts': _TIConst.DsblEnbl.Enbl,
                'Src-Sts': tdev.source_options.index('RmpBO'),
            }

    @property
    def psnames(self):
        """Power supply names."""
        return _dcopy(self._psnames)

    @property
    def psdevices(self):
        """Power supply devices."""
        return self._psdevs

    @property
    def trignames(self):
        """Trigger names."""
        return _dcopy(self._trignames)

    @property
    def trigdevices(self):
        """Trigger devices."""
        return self._trigdevs

    def cmd_turn_off(self):
        """Turn off."""
        retval = self.disable_triggers()
        if not retval[0]:
            return retval

        # wait duration of a ramp for PS change opmode
        _time.sleep(0.5)

        # set slowref
        self.set_devices_propty(
            'OpMode-Sel', _PSConst.OpMode.SlowRef, devices=self._psdevs
        )

        # wait for PS change opmode
        retval = self.wait_devices_propty(
            'OpMode-Sts',
            len(self._psdevs)
            * [[_PSConst.States.SlowRef, _PSConst.States.Off]],
            devices=self._psdevs,
            comp=lambda x, y: x in y,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for BO PS to be in OpMode SlowRef '
                'timed out without success! Verify BO PS!'
            )
            return [False, text, retval[1]]

        # set current to zero
        self.set_devices_propty('Current-SP', 0.0, devices=self._psdevs)

        # wait current change to zero
        retval = self.wait_devices_propty(
            'Current-RB',
            0.0,
            devices=self._psdevs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for BO PS to be with current zero '
                'timed out without success! Verify BO PS!'
            )
            return [False, text, retval[1]]

        return True, '', []

    def cmd_turn_on(self):
        """Turn on."""
        # set rmpwfm
        self.set_devices_propty(
            'OpMode-Sel', _PSConst.OpMode.RmpWfm, devices=self._psdevs
        )

        # wait for PS change opmode
        retval = self.wait_devices_propty(
            'OpMode-Sts',
            _PSConst.States.RmpWfm,
            devices=self._psdevs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for BO PS to be in OpMode RmpWfm '
                'timed out without success! Verify BO PS!'
            )
            return [False, text, retval[1]]

        # set WfmUpdateAuto
        self.set_devices_propty(
            'WfmUpdateAuto-Sel', _PSConst.DsblEnbl.Enbl, devices=self._psdevs
        )

        # wait for PS WfmUpdateAuto to be enabled
        retval = self.wait_devices_propty(
            'WfmUpdateAuto-Sts',
            _PSConst.DsblEnbl.Enbl,
            devices=self._psdevs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for BO PS WfmUpdateAuto to be enabled '
                'timed out without success! Verify BO PS!'
            )
            return [False, text, retval[1]]

        # configure trigger source
        values = [dev.source_options.index('RmpBO') for dev in self._trigdevs]
        self.set_devices_propty('Src-Sel', values, devices=self._trigdevs)

        # wait for trigger source to be configured
        retval = self.wait_devices_propty(
            'Src-Sts',
            values,
            devices=self._trigdevs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for BO Mags Triggers to be in RmpBO event '
                'timed out without success! Verify BO Mags Triggers!'
            )
            return [False, text, retval[1]]

        retval = self.enable_triggers()
        if not retval[0]:
            return retval

        return True, '', []

    def disable_triggers(self):
        """."""
        # disable triggers
        self.set_devices_propty(
            'State-Sel', _TIConst.DsblEnbl.Dsbl, devices=self._trigdevs
        )

        # wait for triggers to be disabled
        retval = self.wait_devices_propty(
            'State-Sts',
            _TIConst.DsblEnbl.Dsbl,
            devices=self._trigdevs,
            timeout=3,
            return_prob=True,
        )

        if retval[0]:
            return True, '', []
        text = (
            'Check for BO Mags Triggers to be disabled timed '
            'out without success! Verify BO Mags Triggers!'
        )
        return [False, text, retval[1]]

    def enable_triggers(self):
        """."""
        # enable triggers
        self.set_devices_propty(
            'State-Sel', _TIConst.DsblEnbl.Enbl, devices=self._trigdevs
        )

        # wait for triggers to be enabled
        retval = self.wait_devices_propty(
            'State-Sts',
            _TIConst.DsblEnbl.Enbl,
            devices=self._trigdevs,
            timeout=3,
            return_prob=True,
        )

        if retval[0]:
            return True, '', []
        text = (
            'Check for BO Mags Triggers to be enabled timed '
            'out without success! Verify BO Mags Triggers!'
        )
        return [False, text, retval[1]]


class BORFRampStandbyHandler(_BaseHandler):
    """BO RF Ramp standby mode handler for injection procedure."""

    def __init__(self, hltiming=None):
        """."""
        self.llrf = ASLLRF(ASLLRF.DEVICES.BO)

        # call base class constructor
        super().__init__((self.llrf,), hltiming=hltiming)

        self.rmptrg = self._hltiming.triggers['BO-Glob:TI-LLRF-Rmp']

        self._on_values = dict()
        self._on_values[self.llrf] = {
            'RmpEnbl-Sts': _Const.DsblEnbl.Enbl,
            'RmpReady-Mon': _Const.DsblEnbl.Enbl,
        }
        self._on_values[self.rmptrg] = {'State-Sts': _Const.DsblEnbl.Enbl}

    def cmd_turn_off(self):
        """Turn off."""
        # set RF ramp to disable
        self.llrf.rmp_enable = _Const.DsblEnbl.Dsbl

        # wait for RF ramp to be disable
        retval = self.wait_devices_propty(
            'RmpEnbl-Sts',
            _Const.DsblEnbl.Dsbl,
            devices=self.llrf,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for BO RF Ramp to be disabled timed '
                'out without success! Verify BO RF Ramp!'
            )
            return [False, text, retval[1]]

        retval = self.disable_triggers()
        if not retval[0]:
            return retval

        return True, '', []

    def cmd_turn_on(self):
        """Turn on."""
        retval = self.enable_triggers()
        if not retval[0]:
            return retval

        # set RF ramp to enabled
        self.llrf.rmp_enable = _Const.DsblEnbl.Enbl

        # wait for RF ramp to be enabled
        retval = self.wait_devices_propty(
            'RmpEnbl-Sts',
            _Const.DsblEnbl.Enbl,
            devices=self.llrf,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for BO RF Ramp to be enabled timed '
                'out without success! Verify BO RF Ramp!'
            )
            return [False, text, retval[1]]

        return True, '', []

    def disable_triggers(self):
        """."""
        # disable triggers
        self.set_devices_propty(
            'State-Sel', _TIConst.DsblEnbl.Dsbl, devices=self.rmptrg
        )

        # wait for triggers to be disabled
        retval = self.wait_devices_propty(
            'State-Sts',
            _TIConst.DsblEnbl.Dsbl,
            devices=self.rmptrg,
            timeout=3,
            return_prob=True,
        )

        if retval[0]:
            return True, '', []
        text = (
            'Check for BO RF Triggers to be disabled timed '
            'out without success! Verify BO RF Triggers!'
        )
        return [False, text, retval[1]]

    def enable_triggers(self):
        """."""
        # enable triggers
        self.set_devices_propty(
            'State-Sel', _TIConst.DsblEnbl.Enbl, devices=self.rmptrg
        )

        # wait for triggers to be enabled
        retval = self.wait_devices_propty(
            'State-Sts',
            _TIConst.DsblEnbl.Enbl,
            devices=self.rmptrg,
            timeout=3,
            return_prob=True,
        )

        if retval[0]:
            return True, '', []
        text = (
            'Check for BO RF Triggers to be enabled timed '
            'out without success! Verify BO RF Triggers!'
        )
        return [False, text, retval[1]]


class LinacStandbyHandler(_BaseHandler):
    """LI LLRF standby mode handler for injection procedure."""

    WAIT_2_TURNON = 2  # [s]
    WAIT_4_PULSES = 1.0  # [s]
    DEF_TRIGS_TIMEOUT = 4.0  # [s]

    def __init__(self, hltiming=None):
        """Init."""
        # create devices
        self._llrf_devs = list()
        for dev in DevLILLRF.DEVICES.ALL:
            self._llrf_devs.append(DevLILLRF(dev))

        self._modnames = LIModltr.DEVICES.ALL

        # modulator devices
        self._moddevs = [LIModltr(mod) for mod in self._modnames]
        self._limps = _Device(
            'LA-CN:H1MPS-1',
            props2init=(
                'Mod1State_I',
                'Mod1State_L',
                'Mod1State_R',
                'Mod2State_I',
                'Mod2State_L',
                'Mod2State_R',
            ),
        )

        devices = self._llrf_devs + self._moddevs
        devices.append(self._limps)

        # call base class constructor
        super().__init__(tuple(devices), hltiming=hltiming)

        # NOTE: Triggers ordering is important so that LINAC LLRF does
        # not have any transient during source migration between events
        # Linac and RmpBO.
        self._trig_names_mod = HLTimeSearch.get_hl_triggers({
            'sec': 'LI',
            'dev': '(Mod|Osc)',
        })
        self._trig_names_ssa = HLTimeSearch.get_hl_triggers({
            'sec': 'LI',
            'dev': 'SSAmp',
        })
        self._trig_names_llrf = HLTimeSearch.get_hl_triggers({
            'sec': 'LI',
            'dev': 'LLRF',
        })
        self._trig_names = (
            self._trig_names_mod + self._trig_names_ssa + self._trig_names_llrf
        )

        self._on_values = dict()
        for dev in self._llrf_devs:
            self._on_values[dev] = {
                'GET_INTEGRAL_ENABLE': _Const.DsblEnbl.Enbl,
                'GET_FB_MODE': _Const.DsblEnbl.Enbl,
            }
        for mdev in self._moddevs:
            self._on_values[mdev] = {
                'CHARGE': _TIConst.DsblEnbl.Enbl,
                'TRIGOUT': _TIConst.DsblEnbl.Enbl,
                'CPS_ALL': 1,
            }
        self._on_values[self._limps] = {
            'Mod1State_I': 0,
            'Mod1State_L': 0,
            'Mod2State_I': 0,
            'Mod2State_L': 0,
        }

    @property
    def modnames(self):
        """Modulator names."""
        return _dcopy(self._modnames)

    @property
    def moddevices(self):
        """Modulator devices."""
        return self._moddevs

    def cmd_turn_off(self):
        """Turn off."""
        # turn feedback off
        self.set_devices_propty(
            'SET_FB_MODE', _Const.DsblEnbl.Dsbl, devices=self._llrf_devs
        )
        # wait for feedback to turn off
        retval = self.wait_devices_propty(
            'GET_FB_MODE',
            _Const.DsblEnbl.Dsbl,
            devices=self._llrf_devs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for LI LLRF Feedback Mode to be off '
                'timed out without success! Verify LI LLRF!'
            )
            return [False, text, retval[1]]

        # turn integral off
        self.set_devices_propty(
            'SET_INTEGRAL_ENABLE',
            _Const.DsblEnbl.Dsbl,
            devices=self._llrf_devs,
        )
        # wait for integral to turn off
        retval = self.wait_devices_propty(
            'GET_INTEGRAL_ENABLE',
            _Const.DsblEnbl.Dsbl,
            devices=self._llrf_devs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for LI LLRF Integral Mode to be off '
                'timed out without success! Verify LI LLRF!'
            )
            return [False, text, retval[1]]

        result = [True, '', []]
        if not self.change_trigs_to_linac_evt()[1]:
            result = [True, 'Some trigger sources did not change.', []]

        # turn modulator trigout off
        self.set_devices_propty(
            'TRIGOUT', _Const.DsblEnbl.Dsbl, devices=self._moddevs
        )
        # wait for modulator trigout to turn off
        retval = self.wait_devices_propty(
            'TRIGOUT',
            _Const.DsblEnbl.Dsbl,
            devices=self._moddevs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for LI modulator TrigOut to be off timed '
                'out without success! Verify LI Modulators!'
            )
            return [False, text, retval[1]]

        # wait for modulators trig.out
        _time.sleep(1)

        # turn modulator charge off
        self.set_devices_propty(
            'CHARGE', _Const.DsblEnbl.Dsbl, devices=self._moddevs
        )
        # wait for modulator charge to turn off
        retval = self.wait_devices_propty(
            'CHARGE',
            _Const.DsblEnbl.Dsbl,
            devices=self._moddevs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for LI modulator Charge to be off timed '
                'out without success! Verify LI Modulators!'
            )
            return [False, text, retval[1]]

        return result

    def cmd_turn_on(self):
        """Turn on."""
        # turn modulators charge on
        self.set_devices_propty(
            'CHARGE', _Const.DsblEnbl.Enbl, devices=self._moddevs
        )
        # wait for modulators charge to turn on
        retval = self.wait_devices_propty(
            'CHARGE',
            _Const.DsblEnbl.Enbl,
            devices=self._moddevs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for LI modulator Charge to be on timed '
                'out without success! Verify LI Modulators!'
            )
            return [False, text, retval[1]]

        # reset modulator
        for dev in self._moddevs:
            if not dev.cmd_reset():
                text = 'Could not reset LI modulator! Verify LI Modulators!'
                return [False, text, [dev.devname + ':RESET']]

        # wait a moment for the modulators charge
        _time.sleep(1)

        # turn modulator trigout on
        self.set_devices_propty(
            'TRIGOUT', _Const.DsblEnbl.Enbl, devices=self._moddevs
        )
        # wait for modulator trigout to turn on
        retval = self.wait_devices_propty(
            'TRIGOUT',
            _Const.DsblEnbl.Enbl,
            devices=self._moddevs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for LI modulator TrigOut to be on timed '
                'out without success! Verify LI Modulators!'
            )
            return [False, text, retval[1]]

        retval = self.check_mps_status()
        if not retval[0]:
            return retval
        retval = self.reset_mps_status()
        if not retval[0]:
            return retval

        result = [True, '', []]
        if not self.change_trigs_to_rmpbo_evt()[1]:
            result = [True, 'Some trigger sources did not change.', []]

        # wait for some pulses
        _time.sleep(LinacStandbyHandler.WAIT_2_TURNON)

        # turn integral on
        self.set_devices_propty(
            'SET_INTEGRAL_ENABLE',
            _Const.DsblEnbl.Enbl,
            devices=self._llrf_devs,
        )
        # wait for integral to turn on
        retval = self.wait_devices_propty(
            'GET_INTEGRAL_ENABLE',
            _Const.DsblEnbl.Enbl,
            devices=self._llrf_devs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for LI LLRF Integral Mode to be on '
                'timed out without success! Verify LI LLRF!'
            )
            return [False, text, retval[1]]

        # turn feedback on
        self.set_devices_propty(
            'SET_FB_MODE', _Const.DsblEnbl.Enbl, devices=self._llrf_devs
        )
        # wait for feedback to turn on
        retval = self.wait_devices_propty(
            'GET_FB_MODE',
            _Const.DsblEnbl.Enbl,
            devices=self._llrf_devs,
            timeout=3,
            return_prob=True,
        )
        if not retval[0]:
            text = (
                'Check for LI LLRF Feedback Mode to be on '
                'timed out without success! Verify LI LLRF!'
            )
            return [False, text, retval[1]]

        return result

    def change_trigs_to_linac_evt(self):
        """Change triggers source to Linac."""
        ret = list()
        # stop triggers to LLRF first
        ret.append(
            self.hltiming.change_triggers_source(
                self._trig_names_llrf,
                new_src='Linac',
                timeout=LinacStandbyHandler.DEF_TRIGS_TIMEOUT,
            )
        )
        _time.sleep(LinacStandbyHandler.WAIT_4_PULSES)  # wait for some pulses
        # stop triggers to SSAmp and Modulators
        ret.append(
            self.hltiming.change_triggers_source(
                self._trig_names_ssa,
                new_src='Linac',
                timeout=LinacStandbyHandler.DEF_TRIGS_TIMEOUT,
            )
        )
        ret.append(
            self.hltiming.change_triggers_source(
                self._trig_names_mod,
                new_src='Linac',
                timeout=LinacStandbyHandler.DEF_TRIGS_TIMEOUT,
            )
        )
        notchanged = list()
        [notchanged.extend(r[0]) for r in ret]
        isok = all([r[1] for r in ret])
        return notchanged, isok

    def change_trigs_to_rmpbo_evt(self):
        """Change triggers source to RmpBO."""
        ret = list()
        # start triggers to SSAmp and Modulators
        ret.append(
            self.hltiming.change_triggers_source(
                self._trig_names_mod,
                new_src='RmpBO',
                timeout=LinacStandbyHandler.DEF_TRIGS_TIMEOUT,
            )
        )
        ret.append(
            self.hltiming.change_triggers_source(
                self._trig_names_ssa,
                new_src='RmpBO',
                timeout=LinacStandbyHandler.DEF_TRIGS_TIMEOUT,
            )
        )
        # start triggers to LLRF after all
        _time.sleep(LinacStandbyHandler.WAIT_4_PULSES)  # wait for some pulses
        ret.append(
            self.hltiming.change_triggers_source(
                self._trig_names_llrf,
                new_src='RmpBO',
                timeout=LinacStandbyHandler.DEF_TRIGS_TIMEOUT,
            )
        )
        notchanged = list()
        [notchanged.extend(r[0]) for r in ret]
        isok = all([r[1] for r in ret])
        return notchanged, isok

    def check_mps_status(self):
        """."""
        # check if mps status is ok
        _t0 = _time.time()
        while _time.time() - _t0 < 5:
            if (
                not self._limps['Mod1State_I']
                and not self._limps['Mod2State_I']
            ):
                return True, '', []

        problems = [
            self._limps.pv_object('Mod' + i + 'State_I').pvname
            for i in ['1', '2']
            if self._limps['Mod' + i + 'State_I']
        ]
        text = (
            'Check for LI modulators MPS Status to be ok timed '
            'out without success! Verify LI Modulators MPS!'
        )
        return [False, text, problems]

    def reset_mps_status(self):
        """."""
        # reset linac mps modulator signal
        self._limps['Mod1State_R'] = 1
        self._limps['Mod2State_R'] = 1
        _time.sleep(1)
        self._limps['Mod1State_R'] = 0
        self._limps['Mod2State_R'] = 0

        # check if mps latch is ok
        _t0 = _time.time()
        while _time.time() - _t0 < 5:
            if (
                not self._limps['Mod1State_L']
                and not self._limps['Mod2State_L']
            ):
                return True, '', []

        problems = [
            self._limps.pv_object('Mod' + i + 'State_L').pvname
            for i in ['1', '2']
            if self._limps['Mod' + i + 'State_L']
        ]
        text = (
            'Check for LI modulators MPS Latch to be ok timed '
            'out without success! Verify LI Modulators MPS!'
        )
        return [False, text, problems]


class InjSysStandbyHandler(_DeviceSet):
    """Injection system standy mode handler."""

    DEF_ON_ORDER = _InjConst.INJSYS_DEF_ON_ORDER
    DEF_OFF_ORDER = _InjConst.INJSYS_DEF_OFF_ORDER
    HANDLER_DESC = {
        'as_pu': 'AS PU (All Septa and Kickers)',
        'bo_ps': 'BO PS Ramp',
        'bo_rf': 'BO RF Ramp',
        'li_rf': 'LINAC (SHB, Klystrons and Modulators)',
    }

    def __init__(self, hltiming=None):
        """Init."""
        self._hltiming = hltiming or HLTiming()
        devs = {
            'as_pu': PUMagsStandbyHandler(hltiming=self._hltiming),
            'bo_ps': BOPSRampStandbyHandler(hltiming=self._hltiming),
            'bo_rf': BORFRampStandbyHandler(hltiming=self._hltiming),
            'li_rf': LinacStandbyHandler(hltiming=self._hltiming),
        }
        self._dev_refs = devs
        self._on_order = InjSysStandbyHandler.DEF_ON_ORDER
        self._off_order = InjSysStandbyHandler.DEF_OFF_ORDER
        self._is_running = ''
        self._done = list()
        self._result = None
        self._thread = None
        self._abort = False

        self._on_values = dict()
        for dev in devs.values():
            self._on_values.update(dev.on_values)

        # call super init
        super().__init__(tuple(devs.values()))

    @property
    def hltiming(self):
        """."""
        return self._hltiming

    @hltiming.setter
    def hltiming(self, hltiming):
        """."""
        if isinstance(hltiming, HLTiming):
            self._hltiming = hltiming

    @property
    def handlers(self):
        """."""
        return self._dev_refs

    @property
    def on_order(self):
        """On setpoints order."""
        return _dcopy(self._on_order)

    @on_order.setter
    def on_order(self, value):
        self._on_order = value

    @property
    def off_order(self):
        """Off setpoints order."""
        return _dcopy(self._off_order)

    @off_order.setter
    def off_order(self, value):
        self._off_order = value

    @property
    def is_on(self):
        """Is on."""
        for dev in self.devices:
            if not dev.is_on:
                return False
        return True

    @property
    def on_values(self):
        """Dict with PVs to be monitored and their respective 'on' value."""
        return self._on_values

    @property
    def is_running(self):
        """Is running."""
        return self._is_running

    @property
    def done(self):
        """Return commands already done."""
        return self._done

    @property
    def result(self):
        """Command result."""
        return self._result

    def cmd_abort(self):
        """Abort command."""
        if self._is_running:
            self._abort = True
        return True

    def cmd_reset_comm_order(self):
        """Reset commands order."""
        self.on_order = InjSysStandbyHandler.DEF_ON_ORDER
        self.off_order = InjSysStandbyHandler.DEF_OFF_ORDER
        return True

    def cmd_turn_on(self, run_in_thread=False):
        """Turn on."""
        return self._command_base('on', run_in_thread)

    def cmd_turn_off(self, run_in_thread=False):
        """Turn off."""
        return self._command_base('off', run_in_thread)

    def get_dev_state(self, devnames):
        """Return the state, on (True) or off (False), for each device in
        devnames.

        Parameters
        ----------
        devnames: list [str]
            A list of strings with the names of the handler devices

        Returns
        -------
        states: list [bool]
            A list of booleans with the state of each handler devices
        """
        states = list()
        for devname in devnames:
            dev = self._dev_refs[devname]
            states.append(dev.is_on)
        return states

    # --- private methods ---

    def _command_base(self, cmmtype, run_in_thread):
        if self.is_running:
            _log.error('Commands already running!')
            return

        if run_in_thread:
            self._is_running = cmmtype
            self._thread = _Thread(
                target=self._command_target, args=[cmmtype], daemon=True
            )
            self._thread.start()
            return

        self._command_target(cmmtype)
        return self._result

    def _command_target(self, cmdtype):
        order = getattr(self, cmdtype + '_order')

        self._result = None
        self._done = list()
        retval = None
        for handler_name in order:
            if self._abort:
                self._abort = False
                break
            handler = self._dev_refs[handler_name]
            func = getattr(handler, 'cmd_turn_' + cmdtype)
            retval = func()
            if retval[0]:
                self._done.append(handler_name)
            else:
                if self._result is None:
                    self._result = retval
                else:
                    self._result[1] += '\n' + retval[1]
                    self._result[2] += retval[2]

        if self._result is None:
            self._result = [True, '', []]

        self._is_running = ''


class InjSysPUModeHandler(_DeviceSet, _Callback):
    """Device to control pulsed magnets configuration for injection."""

    _DEF_TIMEOUT = 10  # [s]
    _DEF_SLEEP = 0.1  # [s]
    SI_DPKCKR_DEFKICK = -6.7  # [mrad]
    TS_POSANG_DEFDELTA = 2.5  # [mrad]
    SI_DPKCKR_DLYREF = 36800000  # [count]

    def __init__(self, print_log=True, callback=None, hltiming=None):
        """Init."""
        self._hltiming = hltiming or HLTiming()
        self.pudpk = PowerSupplyPU(PowerSupplyPU.DEVICES.SI_INJ_DPKCKR)
        self.punlk = PowerSupplyPU(PowerSupplyPU.DEVICES.SI_INJ_NLKCKR)
        self.trigdpk = self._hltiming.triggers['SI-01SA:TI-InjDpKckr']
        self.trignlk = self._hltiming.triggers['SI-01SA:TI-InjNLKckr']
        self.posang = PosAng(PosAng.DEVICES.TS)
        self.delta_posang = self.TS_POSANG_DEFDELTA
        self.dpkckr_dlyref = self.SI_DPKCKR_DLYREF
        self.dpkckr_kick = self.SI_DPKCKR_DEFKICK
        devices = (self.pudpk, self.punlk, self._hltiming, self.posang)
        self._print_log = print_log
        self._abort = _Flag()

        # call super init
        _DeviceSet.__init__(self, devices)
        _Callback.__init__(self, callback=callback)

    @property
    def hltiming(self):
        """."""
        return self._hltiming

    @hltiming.setter
    def hltiming(self, hltiming):
        """."""
        if isinstance(hltiming, HLTiming):
            self._hltiming = hltiming

    @property
    def is_trigdpk_onaxis(self):
        """Whether DpK trigger delay raw is above dpkckr_dlyref."""
        return self.trigdpk.delay_raw > self.dpkckr_dlyref

    @property
    def is_accum(self):
        """If configuration is Accumulation."""
        if not self.connected:
            return False
        dpk_off = (self.pudpk.pwrstate == PowerSupplyPU.PWRSTATE.Off) or (
            self.pudpk.pulse == PowerSupplyPU.PULSTATE.Off
        )
        nlk_on = (self.punlk.pwrstate == PowerSupplyPU.PWRSTATE.On) and (
            self.punlk.pulse == PowerSupplyPU.PULSTATE.On
        )
        return dpk_off & nlk_on

    def cmd_switch_to_accum(self):
        """Switch to Accumulation."""
        self._abort.clear()
        self._update_status('Switching to PU Accumulation config...')
        if not self._check_pu_interlocks():
            return False
        if self._check_abort():
            return False

        # if previously in on-axis, do delta angle x
        if self.is_onaxis:
            if not self._do_delta_posang(-self.delta_posang):
                return False

        # configure DpK trigger
        if not self.is_trigdpk_onaxis:
            delay = self.trigdpk.delay_raw + self.dpkckr_dlyref
            if not self._config_dpk_trigger(delayraw=delay):
                return False

        # set pulsed magnet pwrstate and pulse
        proced = (
            (self.pudpk.cmd_turn_off_pulse, 'turn DpK pulse off.'),
            (self.punlk.cmd_turn_on, 'turn NLK on.'),
            (self.punlk.cmd_turn_on_pulse, 'turn NLK pulse on.'),
        )
        if not self._do_procedures(proced):
            return False

        self._update_status('PU mode switched to Accumulation!')
        return True

    @property
    def is_optim(self):
        """If configuration is Optimization."""
        if not self.connected:
            return False
        dpk_evt_ok = self.trigdpk.source == self.trignlk.source
        dpk_dly_ok = not self.is_trigdpk_onaxis
        dpk_kck_ok = abs(self.pudpk.strength - self.dpkckr_kick) < 1e-3
        dpk_on = (self.pudpk.pwrstate == PowerSupplyPU.PWRSTATE.On) and (
            self.pudpk.pulse == PowerSupplyPU.PULSTATE.On
        )
        nlk_on = (self.punlk.pwrstate == PowerSupplyPU.PWRSTATE.On) and (
            self.punlk.pulse == PowerSupplyPU.PULSTATE.On
        )
        return dpk_evt_ok & dpk_dly_ok & dpk_kck_ok & dpk_on & nlk_on

    def cmd_switch_to_optim(self):
        """Switch to Optimization."""
        self._abort.clear()
        self._update_status('Switching to PU Optimization config...')
        if not self._check_pu_interlocks():
            return False
        if self._check_abort():
            return False

        # if previously in on-axis, do delta angle x
        if self.is_onaxis:
            if not self._do_delta_posang(-self.delta_posang):
                return False

        if self._check_abort():
            return False

        # configure DpK trigger
        if self.is_trigdpk_onaxis:
            delay = self.trigdpk.delay_raw - self.dpkckr_dlyref
            if not self._config_dpk_trigger(delayraw=delay):
                return False

        # set DpK Kick
        if not self._config_dpk_kick():
            return False

        # set pulsed magnet pwrstate and pulse
        proced = (
            (self.pudpk.cmd_turn_on, 'turn DpK on.'),
            (self.punlk.cmd_turn_on, 'turn NLK on.'),
            (self.pudpk.cmd_turn_on_pulse, 'turn DpK pulse on.'),
            (self.punlk.cmd_turn_on_pulse, 'turn NLK pulse on.'),
        )
        if not self._do_procedures(proced):
            return False

        self._update_status('PU mode switched to Optimization!')
        return True

    @property
    def is_onaxis(self):
        """If configuration is OnAxis."""
        if not self.connected:
            return False
        dpk_evt_ok = self.trigdpk.source == self.trignlk.source
        dpk_dly_ok = self.is_trigdpk_onaxis
        dpk_kck_ok = abs(self.pudpk.strength - self.dpkckr_kick) < 1e-3
        dpk_on = (self.pudpk.pwrstate == PowerSupplyPU.PWRSTATE.On) and (
            self.pudpk.pulse == PowerSupplyPU.PULSTATE.On
        )
        nlk_off = (self.punlk.pwrstate == PowerSupplyPU.PWRSTATE.Off) or (
            self.punlk.pulse == PowerSupplyPU.PULSTATE.Off
        )
        return dpk_evt_ok & dpk_dly_ok & dpk_kck_ok & dpk_on & nlk_off

    def cmd_switch_to_onaxis(self):
        """Switch to OnAxis."""
        self._abort.clear()
        self._update_status('Switching to PU OnAxis config...')
        if not self._check_pu_interlocks():
            return False
        if self._check_abort():
            return False

        # if not previously in on-axis, do delta angle x
        if not self.is_onaxis:
            if not self._do_delta_posang(self.delta_posang):
                return False

        if self._check_abort():
            return False

        # configure DpK trigger
        if not self.is_trigdpk_onaxis:
            delay = self.trigdpk.delay_raw + self.dpkckr_dlyref
            if not self._config_dpk_trigger(delayraw=delay):
                return False

        # set DpK Kick
        if not self._config_dpk_kick():
            return False

        # set pulsed magnet pwrstate and pulse
        proced = (
            (self.punlk.cmd_turn_off_pulse, 'turn NLK pulse off.'),
            (self.pudpk.cmd_turn_on, 'turn DpK on.'),
            (self.pudpk.cmd_turn_on_pulse, 'turn DpK pulse on.'),
        )
        if not self._do_procedures(proced):
            return False

        self._update_status('PU mode switched to OnAxis!')
        return True

    def _check_pu_interlocks(self):
        if not self.pudpk.interlock_ok:
            self._update_status('ERR: DpK interlocks not ok. Aborted.')
            return False
        if not self.punlk.interlock_ok:
            self._update_status('ERR: NLK interlocks not ok. Aborted.')
            return False
        return True

    def _do_delta_posang(self, delta):
        if self.posang.need_ref_update:
            self.posang.cmd_update_reference()
        _time.sleep(InjSysPUModeHandler._DEF_SLEEP)
        desval = self.posang.delta_angx + delta
        self.posang.delta_angx += delta
        if not self._wait(self.posang, 'delta_angx', desval):
            self._update_status('ERR:Could not do delta AngX.')
            return False
        return True

    def _config_dpk_trigger(self, delayraw):
        # set DpK trigger event to NLK trigger event
        dpk_event = self.trignlk.source
        self.trigdpk.source = dpk_event
        if not self._wait(self.trigdpk, 'source', dpk_event):
            self._update_status('ERR:Could not set DpK trigger event.')
            return False
        # set DpK trigger DelayRaw
        self.trigdpk.delay_raw = delayraw
        if not self._wait(self.trigdpk, 'delay_raw', delayraw):
            self._update_status('ERR:Could not set DpK trigger delay.')
            return False
        return True

    def _config_dpk_kick(self):
        self.pudpk.strength = self.dpkckr_kick
        if not self._wait(self.pudpk, 'strength', self.dpkckr_kick):
            self._update_status(
                f'ERR:Could not set DpK Kick to {self.dpkckr_kick:.1f}mrad.'
            )
            return False
        return True

    def _do_procedures(self, procedures):
        for fun, msg in procedures:
            if self._check_abort():
                return False
            if not fun(timeout=2):
                self._update_status('ERR:Could not ' + msg)
                return False
            _time.sleep(0.5)
        return True

    # -------- thread help methods --------

    def cmd_abort(self):
        """Abort PU mode change."""
        self._abort.set()
        self._update_status('WARN:Abort received for PU Mode change.')
        return True

    def _check_abort(self):
        """Clear abort flag."""
        if not self._abort.is_set():
            return False
        self._abort.clear()
        self._update_status('ERR:Aborted PU Mode change.')
        return True

    # ---------- check sp -----------

    def _wait(
        self, device, prop, desired, tolerance=1e-3, timeout=_DEF_TIMEOUT
    ):
        _t0 = _time.time()
        while _time.time() - _t0 < timeout:
            if abs(getattr(device, prop) - desired) < tolerance:
                return True
            _time.sleep(InjSysPUModeHandler._DEF_SLEEP)
        return False

    # ---------- logging -----------

    def _update_status(self, status):
        if self._print_log:
            print(status)
        self.run_callbacks(status)



"""Insertion Device Feedforward Devices."""

import inspect as _inspect
import time as _time
from copy import deepcopy as _dcopy

import numpy as _np

from ..idff.config import IDFFConfig as _IDFFConfig
from ..idff.csdev import IDFFConst as _IDFFConst
from ..namesys import SiriusPVName as _SiriusPVName
from ..search import IDSearch as _IDSearch
from .device import Device as _Device, DeviceSet as _DeviceSet
from .ids import ID as _ID
from .pwrsupply import PowerSupplyFBP as _PowerSupplyFBP


class _ParamPVs:
    """."""

    LOOPSTATE_SEL = 'LoopState-Sel'
    LOOPSTATE_STS = 'LoopState-Sts'
    LOG_MON = None
    LOOPFREQ_SP = None
    LOOPFREQ_RB = None
    POLARIZATION_MON = None
    CONFIGNAME_SP = None
    CONFIGNAME_RB = None
    CORRCONFIG_CMD = None
    CORRSTATUS_MON = None
    CORRSTATUSLABELS_CTE = None
    CORRCH_1CURRENT_MON = None
    CORRCH_2CURRENT_MON = None
    CORRCV_1CURRENT_MON = None
    CORRCV_2CURRENT_MON = None
    CORRQS_1CURRENT_MON = None
    CORRQS_2CURRENT_MON = None
    CORRLCHCURRENT_MON = None
    CORRLCVCURRENT_MON = None
    CORRQD1_1CURRENT_MON = None
    CORRQF_1CURRENT_MON = None
    CORRQD2_1CURRENT_MON = None
    CORRQD2_2CURRENT_MON = None
    CORRQF_2CURRENT_MON = None
    CORRQD1_2CURRENT_MON = None
    CORRCC1_1CURRENT_MON = None
    CORRCC2_1CURRENT_MON = None
    CORRCC2_2CURRENT_MON = None
    CORRCC1_2CURRENT_MON = None
    TABLE_SP = None
    TABLE_RB = None

    def __str__(self):
        """Print parameters."""
        str_ = ''
        strf = '{}: {}'
        for key, value in _inspect.getmembers(self):
            if not key.startswith('_') and value is not None:
                lstr = strf.format(key, value)
                str_ += lstr if str_ == '' else '\n' + lstr
        return str_


class IDFFCtrlBase(_Device):
    """ID Feedforward Control Device Base."""

    _DEF_TIMEOUT = 2  # [s]

    IDFF_CH_LABELS = tuple()
    IDFF_CV_LABELS = tuple()
    IDFF_QS_LABELS = tuple()
    IDFF_LC_LABELS = tuple()
    IDFF_QN_LABELS = tuple()
    IDFF_CC_LABELS = tuple()

    PARAM_PVS = _ParamPVs()

    PROPERTIES_DEFAULT = tuple(
        set(
            value
            for key, value in _inspect.getmembers(PARAM_PVS)
            if not key.startswith('_') and value is not None
        )
    )

    def __init__(self, devname, props2init='all', auto_monitor_mon=True):
        """."""
        # call base class constructor
        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon
        )

    @property
    def loopstate(self):
        """Return loop state."""
        return self[self.PARAM_PVS.LOOPSTATE_STS]

    @property
    def correctors_status(self):
        """Return correctors status."""
        curr_sts = self.PARAM_PVS.CORRSTATUS_MON
        return self[curr_sts] if curr_sts else None

    @property
    def calc_corr_current_ch_1(self):
        """Return calculated CH_1 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRCH_1CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_ch_2(self):
        """Return calculated CH_2 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRCH_2CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_cv_1(self):
        """Return calculated CV_1 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRCV_1CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_cv_2(self):
        """Return calculated CV_2 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRCV_2CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_qs_1(self):
        """Return calculated QS_1 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRQS_1CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_qs_2(self):
        """Return calculated Q_2 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRQS_2CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_lch(self):
        """Return calculated LCH power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRLCHCURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_lcv(self):
        """Return calculated LCV power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRLCVCURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_qd1_1(self):
        """Return calculated QD1_1 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRQD1_1CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_qf_1(self):
        """Return calculated QF_1 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRQF_1CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_qd2_1(self):
        """Return calculated QD2_1 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRQD2_1CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_qd2_2(self):
        """Return calculated QD2_2 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRQD2_2CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_qf_2(self):
        """Return calculated QF_2 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRQF_2CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_qd1_2(self):
        """Return calculated QD1_2 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRQD1_2CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_cc1_1(self):
        """Return calculated CC1_1 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRCC1_1CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_cc2_1(self):
        """Return calculated CC2_1 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRCC2_1CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_cc2_2(self):
        """Return calculated CC2_2 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRCC2_2CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def calc_corr_current_cc1_2(self):
        """Return calculated CC1_2 power supply current [A]."""
        curr_name = self.PARAM_PVS.CORRCC1_2CURRENT_MON
        return self[curr_name] if curr_name else None

    @property
    def loopfreq(self):
        """Return FF loop frequency [Hz]."""
        loopfreq_rb_name = self.PARAM_PVS.LOOPFREQ_RB
        return self[loopfreq_rb_name] if loopfreq_rb_name else None

    @loopfreq.setter
    def loopfreq(self, value):
        """Set FF loop frequency [Hz]."""
        loopfreq_sp_name = self.PARAM_PVS.LOOPFREQ_SP
        if loopfreq_sp_name is not None:
            self[loopfreq_sp_name] = value

    def cmd_turn_on_feedforward(self, timeout=None):
        """."""
        timeout = timeout or self._DEF_TIMEOUT
        if self.loopstate == _IDFFConst.LoopState.Closed:
            return True
        self['LoopState-Sel'] = _IDFFConst.LoopState.Closed
        return self.wait(
            'LoopState-Sts', _IDFFConst.LoopState.Closed, timeout=timeout
        )

    def cmd_turn_off_feedforward(self, timeout=None):
        """."""
        timeout = timeout or self._DEF_TIMEOUT
        if self.loopstate == _IDFFConst.LoopState.Open:
            return True
        self['LoopState-Sel'] = _IDFFConst.LoopState.Open
        return self.wait(
            'LoopState-Sts', _IDFFConst.LoopState.Open, timeout=timeout
        )

    @staticmethod
    def conv_idffdevname_2_iddevname(idffdevname):
        """."""
        iddevname = _SiriusPVName(idffdevname)
        iddevname = iddevname.substitute(dis='ID', idx='')
        if iddevname.sub in ('08SB', '14SB'):
            iddevname = iddevname.substitute(dev='IVU18')
        elif iddevname.sub in ('06SB', '07SP'):
            iddevname = iddevname.substitute(dev='VPU29')
        elif iddevname.sub in ('10SB',):
            iddevname = iddevname.substitute(dev='DELTA52')
        elif iddevname.sub in ('11SP',):
            iddevname = iddevname.substitute(dev='SIMUL')
        else:
            pass
        return iddevname

    @staticmethod
    def _add_devices(devset_base, devset_derived):
        for key, value in _inspect.getmembers(devset_derived):
            if key == 'ALL':
                alldevs = devset_base.ALL + value
                devset_base.ALL = alldevs
            elif not key.startswith('_'):
                setattr(devset_base, key, value)


class IDFFCtrlSoft(IDFFCtrlBase):
    """ID Feedforward Control Soft Device."""

    class DEVICES:
        """Device names."""

        ALL = tuple()
        # should be added in derived classes

    PARAM_PVS = _dcopy(IDFFCtrlBase.PARAM_PVS)
    PARAM_PVS.LOG_MON = 'Log-Mon'
    PARAM_PVS.LOOPFREQ_SP = 'LoopFreq-SP'
    PARAM_PVS.LOOPFREQ_RB = 'LoopFreq-RB'
    PARAM_PVS.POLARIZATION_MON = 'Polarization-Mon'
    PARAM_PVS.CONFIGNAME_SP = 'ConfigName-SP'
    PARAM_PVS.CONFIGNAME_RB = 'ConfigName-RB'
    PARAM_PVS.CORRCONFIG_CMD = 'CorrConfig-Cmd'
    PARAM_PVS.CORRSTATUS_MON = 'CorrStatus-Mon'
    PARAM_PVS.CORRSTATUSLABELS_CTE = 'CorrStatusLabels-Cte'

    PROPERTIES_DEFAULT = tuple(
        set(
            value
            for key, value in _inspect.getmembers(PARAM_PVS)
            if not key.startswith('_') and value is not None
        )
    )

    @property
    def configname(self):
        """Return configuration name."""
        if self.PARAM_PVS.CONFIGNAME_SP:
            return self[self.PARAM_PVS.CONFIGNAME_RB]


class IDFFCtrlSoftDELTA(IDFFCtrlSoft):
    """ID Feedforward Control Soft Device for DELTA."""

    class DEVICES:
        """Device names."""

        DELTA52_10SB_SOFT = 'SI-10SB:AP-IDFF'
        ALL = (DELTA52_10SB_SOFT,)

    IDFFCtrlBase._add_devices(IDFFCtrlSoft.DEVICES, DEVICES)

    IDFF_CH_LABELS = _IDSearch.IDFF_CH_LABELS
    IDFF_CV_LABELS = _IDSearch.IDFF_CV_LABELS
    IDFF_QS_LABELS = _IDSearch.IDFF_QS_LABELS

    PARAM_PVS = _dcopy(IDFFCtrlSoft.PARAM_PVS)
    PARAM_PVS.CORRCH_1CURRENT_MON = 'CorrCH_1Current-Mon'
    PARAM_PVS.CORRCH_2CURRENT_MON = 'CorrCH_2Current-Mon'
    PARAM_PVS.CORRCV_1CURRENT_MON = 'CorrCV_1Current-Mon'
    PARAM_PVS.CORRCV_2CURRENT_MON = 'CorrCV_2Current-Mon'
    PARAM_PVS.CORRQS_1CURRENT_MON = 'CorrQS_1Current-Mon'
    PARAM_PVS.CORRQS_2CURRENT_MON = 'CorrQS_2Current-Mon'

    PROPERTIES_DEFAULT = tuple(
        set(
            value
            for key, value in _inspect.getmembers(PARAM_PVS)
            if not key.startswith('_') and value is not None
        )
    )

    def __init__(self, devname=None, props2init='all', auto_monitor_mon=True):
        """."""
        # check if device exists
        if devname is None:
            devname = self.DEVICES.DELTA52_10SB_SOFT
        if devname not in self.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon
        )


class IDFFCtrlSoftIVU(IDFFCtrlSoft):
    """ID Feedforward Control Soft Device for IVU."""

    class DEVICES:
        """Device names."""

        IVU18_08SB_SOFT = 'SI-08SB:AP-IDFF'
        IVU18_14SB_SOFT = 'SI-14SB:AP-IDFF'
        ALL = (IVU18_08SB_SOFT, IVU18_14SB_SOFT)

    IDFFCtrlBase._add_devices(IDFFCtrlSoft.DEVICES, DEVICES)

    # IDFF_LC_LABELS = _IDSearch.IDFF_LC_LABELS
    IDFF_QN_LABELS = _IDSearch.IDFF_QN_LABELS

    PARAM_PVS = _dcopy(IDFFCtrlSoft.PARAM_PVS)
    PARAM_PVS.CORRQD1_1CURRENT_MON = 'CorrQD1_1Current-Mon'
    PARAM_PVS.CORRQF_1CURRENT_MON = 'CorrQF_1Current-Mon'
    PARAM_PVS.CORRQD2_1CURRENT_MON = 'CorrQD2_1Current-Mon'
    PARAM_PVS.CORRQD1_2CURRENT_MON = 'CorrQD1_2Current-Mon'
    PARAM_PVS.CORRQF_2CURRENT_MON = 'CorrQF_2Current-Mon'
    PARAM_PVS.CORRQD2_2CURRENT_MON = 'CorrQD2_2Current-Mon'

    PROPERTIES_DEFAULT = tuple(
        set(
            value
            for key, value in _inspect.getmembers(PARAM_PVS)
            if not key.startswith('_') and value is not None
        )
    )

    def __init__(self, devname=None, props2init='all', auto_monitor_mon=True):
        """."""
        # check if device exists
        if devname not in self.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon
        )


class IDFFCtrlSoftVPU(IDFFCtrlSoft):
    """ID Feedforward Control Soft Device for VPU."""

    class DEVICES:
        """Device names."""

        VPU29_06SB_SOFT = 'SI-06SB:AP-IDFF'
        VPU29_07SP_SOFT = 'SI-07SP:AP-IDFF'
        ALL = (VPU29_06SB_SOFT, VPU29_07SP_SOFT)

    IDFFCtrlBase._add_devices(IDFFCtrlSoft.DEVICES, DEVICES)

    # IDFF_LC_LABELS = _IDSearch.IDFF_LC_LABELS
    IDFF_CC_LABELS = _IDSearch.IDFF_CC_LABELS

    PARAM_PVS = IDFFCtrlSoft.PARAM_PVS
    PARAM_PVS.CORRCC1_1CURRENT_MON = 'CorrCC1_1Current-Mon'
    PARAM_PVS.CORRCC2_1CURRENT_MON = 'CorrCC2_1Current-Mon'
    PARAM_PVS.CORRCC1_2CURRENT_MON = 'CorrCC1_2Current-Mon'
    PARAM_PVS.CORRCC2_2CURRENT_MON = 'CorrCC2_2Current-Mon'

    PROPERTIES_DEFAULT = tuple(
        set(
            value
            for key, value in _inspect.getmembers(PARAM_PVS)
            if not key.startswith('_') and value is not None
        )
    )

    def __init__(self, devname=None, props2init='all', auto_monitor_mon=True):
        """."""
        # check if device exists
        if devname not in self.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon
        )


class IDFFCtrlHard(IDFFCtrlBase):
    """ID Feedforward Control Hard Device."""

    class DEVICES:
        """Device names."""

        ALL = tuple()
        # should be added in derived classes

    PARAM_PVS = _dcopy(IDFFCtrlBase.PARAM_PVS)
    PARAM_PVS.TABLE_SP = 'Table-SP'
    PARAM_PVS.TABLE_RB = 'Table-RB'

    PROPERTIES_DEFAULT = tuple(
        set(
            value
            for key, value in _inspect.getmembers(PARAM_PVS)
            if not key.startswith('_') and value is not None
        )
    )

    def get_ffwd_table_corr_labels(self):
        """."""
        corr_labels = list()
        corr_labels += self.IDFF_CH_LABELS
        corr_labels += self.IDFF_CV_LABELS
        corr_labels += self.IDFF_CC_LABELS
        corr_labels += self.IDFF_LC_LABELS
        corr_labels += self.IDFF_QS_LABELS
        corr_labels += self.IDFF_QN_LABELS
        return corr_labels

    def get_ffwd_table(self):
        """Return FF table dict."""
        param_name = self.PARAM_PVS.TABLE_RB
        if param_name is None:
            return dict()
        ff_table = _np.array(self[param_name])
        clabels = self.get_ffwd_table_corr_labels()
        ff_table = ff_table.reshape(len(clabels), -1)
        ff_table = {clabels[i]: ff_table[i, :] for i in range(len(clabels))}
        return ff_table


class IDFFCtrlHardIVU(IDFFCtrlHard):
    """ID Feedforward Control IVU Device."""

    class DEVICES:
        """Device names."""

        IVU18_08SB_HARD = 'SI-08SB:BS-IDFF-CHCV'
        IVU18_11SP_HARD = 'SI-11SP:BS-IDFF-CHCV'
        IVU18_14SB_HARD = 'SI-14SB:BS-IDFF-CHCV'
        ALL = (IVU18_08SB_HARD, IVU18_11SP_HARD, IVU18_14SB_HARD)

    IDFFCtrlBase._add_devices(IDFFCtrlHard.DEVICES, DEVICES)

    IDFF_CH_LABELS = _IDSearch.IDFF_CH_LABELS
    IDFF_CV_LABELS = _IDSearch.IDFF_CV_LABELS


class IDFFCtrlHardVPU(IDFFCtrlHard):
    """ID Feedforward Control VPU Device."""

    class DEVICES:
        """Device names."""

        VPU29_06SB_HARD = 'SI-06SB:BS-IDFF-CC'
        VPU29_07SP_HARD = 'SI-07SP:BS-IDFF-CC'
        ALL = (VPU29_06SB_HARD, VPU29_07SP_HARD)

    IDFFCtrlBase._add_devices(IDFFCtrlHard.DEVICES, DEVICES)

    IDFF_CC_LABELS = _IDSearch.IDFF_CC_LABELS


class IDFFCtrl(IDFFCtrlBase):
    """ID Feedforward Control Device."""

    class DEVICES:
        """Device names."""

        IDFFCtrlSoft = IDFFCtrlSoft.DEVICES
        IDFFCtrlHard = IDFFCtrlHard.DEVICES
        ALL = IDFFCtrlSoft.ALL + IDFFCtrlHard.ALL

    def __new__(cls, devname, **kwargs):
        """."""
        idclass = IDFFCtrl.get_idffclass(devname)
        if idclass:
            return idclass(devname, **kwargs)
        else:
            raise NotImplementedError(devname)

    @staticmethod
    def get_idffclass(devname):
        """."""
        if devname in IDFFCtrlSoftDELTA.DEVICES.ALL:
            return IDFFCtrlSoftDELTA
        if devname in IDFFCtrlSoftIVU.DEVICES.ALL:
            return IDFFCtrlSoftIVU
        elif devname in IDFFCtrlHardIVU.DEVICES.ALL:
            return IDFFCtrlHardIVU
        if devname in IDFFCtrlSoftVPU.DEVICES.ALL:
            return IDFFCtrlSoftVPU
        elif devname in IDFFCtrlHardVPU.DEVICES.ALL:
            return IDFFCtrlHardVPU
        return None


class IDFF(_DeviceSet):
    """ID Feedforward System Device."""

    class DEVICES:
        """Device names."""

        ALL = tuple()

    IDFFCtrlBase._add_devices(DEVICES, IDFFCtrl.DEVICES)

    def __init__(
        self,
        devname,
        props2init_ctrl='all',
        props2init_corrs='all',
        with_devctrl=True,
    ):
        """."""
        self._with_devctrl = with_devctrl

        # check if device exists
        if devname not in IDFF.DEVICES.ALL:
            raise NotImplementedError(devname)
        self._devname = devname
        self._idffclass = IDFFCtrl.get_idffclass(self._devname)
        self._iddevname = IDFFCtrlBase.conv_idffdevname_2_iddevname(devname)

        # self._devname = iddevname  # needed for _create_devices
        self._idffconfig = _IDFFConfig()

        self._pparametername = _IDSearch.conv_idname_2_pparameter_propty(
            self._iddevname
        )
        self._kparametername = _IDSearch.conv_idname_2_kparameter_propty(
            self._iddevname
        )

        alldevs = self._create_devices(props2init_ctrl, props2init_corrs)
        (
            self._devctrl,
            self._devid,
            self._devsch,
            self._devscv,
            self._devsqs,
            self._devslc,
            self._devsqn,
            self._devscc,
        ) = alldevs

        self._lab2corrdevs = self._create_labels_2_corrdevs_dict()

        devices = list()
        if self._with_devctrl:
            devices += [self._devctrl, self._devid]
        else:
            devices += [self._devid]
        devices += self._devsch
        devices += self._devscv
        devices += self._devsqs
        devices += self._devslc
        devices += self._devsqn
        devices += self._devscc
        super().__init__(devices, devname=devname)

    @property
    def idffclass(self):
        """Return IDFFCtrl class."""
        return self._idffclass

    @property
    def iddevname(self):
        """."""
        return self._iddevname

    @property
    def chnames(self):
        """Return CH corrector power supply names."""
        return self._get_corrnames(self.idffclass.IDFF_CH_LABELS)

    @property
    def cvnames(self):
        """Return CV corrector power supply names."""
        return self._get_corrnames(self.idffclass.IDFF_CV_LABELS)

    @property
    def qsnames(self):
        """Return QS corrector power supply names."""
        return self._get_corrnames(self.idffclass.IDFF_QS_LABELS)

    @property
    def lcnames(self):
        """Return LC corrector power supply names."""
        return self._get_corrnames(self.idffclass.IDFF_LC_LABELS)

    @property
    def qnnames(self):
        """Return QN trim corrector power supply names."""
        return self._get_corrnames(self.idffclass.IDFF_QN_LABELS)

    @property
    def ccnames(self):
        """Return CC corrector power supply names."""
        return self._get_corrnames(self.idffclass.IDFF_CC_LABELS)

    @property
    def ctrldev(self):
        """Return IDFFCtrl device."""
        return self._devctrl

    @property
    def iddev(self):
        """Return ID device."""
        return self._devid

    @property
    def chdevs(self):
        """Return CH corrector power supply devices."""
        return self._devsch

    @property
    def cvdevs(self):
        """Return CV corrector power supply devices."""
        return self._devscv

    @property
    def qsdevs(self):
        """Return QS corrector power supply names."""
        return self._devsqs

    @property
    def lcdevs(self):
        """Return LC corrector power supply names."""
        return self._devslc

    @property
    def qndevs(self):
        """Return QD corrector power supply names."""
        return self._devsqn

    @property
    def ccdevs(self):
        """Return CC corrector power supply names."""
        return self._devscc

    @property
    def pparametername(self):
        """Return corresponding to ID pparameter."""
        return self._pparametername

    @property
    def kparametername(self):
        """Return corresponding to ID kparameter."""
        return self._kparametername

    @property
    def polarizations(self):
        """Return list of possible light polarizations for the ID."""
        return _IDSearch.conv_idname_2_polarizations(self.iddevname)

    @property
    def polarization_mon(self):
        """Return current ID polarization as a string (or None)."""
        return self.iddev.polarization_mon_str

    @property
    def pparameter_mon(self):
        """Return pparameter value."""
        if self._pparametername:
            return self._devid[self._pparametername]
        else:
            return None

    @property
    def kparameter_mon(self):
        """Return kparameter value."""
        return self._devid[self._kparametername]

    @property
    def idffconfig(self):
        """."""
        return self._idffconfig

    def read_corr_offset_values(self):
        """Read current corrector values."""
        offsets = dict()
        for label, corrdev in self._lab2corrdevs.items():
            offsets[label] = corrdev.current
        return offsets

    def find_configs(self):
        """Find si_idff configurations in configdb."""
        return self._idffconfig.configdbclient.find_configs()

    def load_config(self, name):
        """Load IDFF configuration."""
        self._idffconfig.name = name
        if self._idffconfig.name == name:
            self._idffconfig.load()
        else:
            raise ValueError('Could not load configuration.')

    def read_setpoints(self, corrdevs=None):
        """Return corrector SP values."""
        if corrdevs is None:
            corrdevs = self._devsch + self._devscv + self._devsqs
        chs = _IDSearch.conv_idname_2_idff_chnames(self.devname)
        cvs = _IDSearch.conv_idname_2_idff_cvnames(self.devname)
        qss = _IDSearch.conv_idname_2_idff_qsnames(self.devname)
        lcs = _IDSearch.conv_idname_2_idff_lcnames(self.devname)
        qns = _IDSearch.conv_idname_2_idff_qnnames(self.devname)
        ccs = _IDSearch.conv_idname_2_idff_ccnames(self.devname)
        corrs = chs + cvs + qss + lcs + qns + ccs
        setpoints = dict()
        for pvname in corrs:
            # find corrdev corresponding to pvname
            for dev in corrdevs:
                if dev.devname in pvname:
                    spvname = _SiriusPVName(pvname)
                    # propty = spvname.propty.replace('-SP', '-RB')
                    setpoints[pvname] = dev[spvname.propty]
        return setpoints

    def calculate_setpoints(
        self,
        pparameter_value=None,
        kparameter_value=None,
        use_ioc_tables=None,
    ):
        """Return correctors setpoints for a particular ID config.

        polarization - a string defining the required polarization for
        setpoint calculation.
        """
        if use_ioc_tables is None:
            use_ioc_tables = issubclass(self._idffclass, IDFFCtrlHard)

        if use_ioc_tables:
            if kparameter_value is None:
                kparameter_value = self.kparameter_mon

            ff_tables = self.ctrldev.get_ffwd_table()
            setpoints = dict()

            idparams = _IDSearch.conv_idname_2_parameters(self.iddevname)
            idff = _IDSearch.conv_idname_2_idff(self.iddevname)
            for corrlabel, ff_table in ff_tables.items():
                # IOC tables gap zero gap offset!
                klims = 0 * idparams.KPARAM_MIN, idparams.KPARAM_MAX
                kparam = _np.linspace(*klims, len(ff_table))
                # linear interpolation
                curr = _np.interp(kparameter_value, kparam, ff_table)
                corr_pvname = idff[corrlabel]
                setpoints[corr_pvname] = curr
            sts = (
                setpoints, self.polarization_mon,
                self.pparameter_mon, kparameter_value
            )
            return sts

        # NOTE:
        # For standardization soft IDFF IOCs could have Table-(SP|RB) PVs
        # like the hard IDFF IOCs have...
        if not self._idffconfig:
            ValueError('IDFFConfig is not loaded!')

        polarization, pparameter_value, kparameter_value = (
            self.get_polarization_state(
                pparameter_value=pparameter_value,
                kparameter_value=kparameter_value,
            )
        )

        if polarization not in self.idffconfig.polarizations:
            raise ValueError('Polarization is not compatible with ID.')
        if pparameter_value is None:
            pparameter_value = self.pparameter_mon
        if kparameter_value is None:
            kparameter_value = self.kparameter_mon
        setpoints = self.idffconfig.calculate_setpoints(
            polarization, pparameter_value, kparameter_value
        )
        return setpoints, polarization, pparameter_value, kparameter_value

    def implement_setpoints(self, setpoints=None, corrdevs=None):
        """Implement setpoints in correctors."""
        if setpoints is None:
            setpoints, polarization, pparameter_value, kparameter_value = (
                self.calculate_setpoints(
                    pparameter_value=None, kparameter_value=None
                )
            )
        else:
            polarization, pparameter_value, kparameter_value = [None] * 3
        if corrdevs is None:
            corrdevs = (
                self._devsch
                + self._devscv
                + self._devsqs
                + self._devslc
                + self._devsqn
                + self._devscc
            )
        for pvname, value in setpoints.items():
            # find corrdev corresponding to pvname
            for dev in corrdevs:
                if dev.devname in pvname:
                    pvname = _SiriusPVName(pvname)
                    dev[pvname.propty] = value
                    break
        return polarization, pparameter_value, kparameter_value

    def check_valid_value(self, value):
        """Check consistency of SI_IDFF configuration."""
        if not super().check_valid_value(value):
            raise ValueError('Value incompatible with config template')

        configs = value['polarizations']
        pvnames = {
            key: value
            for key, value in value['pvnames']
            if key not in ('pparameters', 'kparameters')
        }
        corrlabels = set(pvnames.keys())

        # check pvnames in configs
        pvsconfig = set(pvnames.values())
        pvsidsearch = set(
            self.chnames
            + self.cvnames
            + self.qsnames
            + self.lcnames
            + self.ccnames
        )
        symm_diff = pvsconfig ^ pvsidsearch
        if symm_diff:
            raise ValueError('List of pvnames in config is not consistent')

        # check polarizations in configs
        pconfig = set(configs.keys())
        pidsearch = set(_IDSearch.conv_idname_2_polarizations(self.devname))
        symm_diff = pconfig ^ pidsearch
        if symm_diff:
            raise ValueError(
                'List of polarizations in config is not consistent'
            )

        # check polarization tables consistency
        for polarization, table in configs.items():
            corrtable = {
                key: value
                for key, value in table
                if key not in ('pparameters', 'kparameters')
            }

            # check 'pparameter'
            if 'pparameter' not in table:
                raise ValueError(
                    'Missing pparameter in polarization configuration.'
                )

            # check 'kparameter'
            if 'kparameter' not in table:
                raise ValueError(
                    'Missing kparameter in polarization configuration.'
                )

            # check corr label list
            corrlabels_config = set(corrtable.keys())
            symm_diff = corrlabels ^ corrlabels_config
            if symm_diff:
                raise ValueError(
                    'List of corrlabels in config is not consistent'
                )

            # check nrpts in tables
            param = 'pparameter' if polarization == 'none' else 'kparameter'
            nrpts_corrtables = {len(table) for table in corrtable.values()}
            nrpts_kparameter = set([len(table[param])])
            symm_diff = nrpts_corrtables ^ nrpts_kparameter
            if symm_diff:
                raise ValueError(
                    'Corrector tables and kparameter list in config '
                    'are not consistent'
                )
        return True

    def get_polarization_state(
        self, pparameter_value=None, kparameter_value=None
    ):
        """."""
        if self.pparametername is None:
            if kparameter_value is None:
                kparameter_value = self.kparameter_mon
            return self.polarizations[0], pparameter_value, kparameter_value
        if pparameter_value is None:
            pparameter_value = self.pparameter_mon
        if kparameter_value is None:
            kparameter_value = self.kparameter_mon
        if None in (pparameter_value, kparameter_value):
            return None, pparameter_value, kparameter_value
        polarization = self.polarization_mon
        if polarization is None:
            polarization = self.idffconfig.get_polarization_state(
                pparameter=pparameter_value, kparameter=kparameter_value
            )
        return polarization, pparameter_value, kparameter_value

    def rampup_corr_currents(
        self,
        nrpts=50,
        time_interval=10,
        pparameter_value=None,
        kparameter_value=None,
        dry_run=False,
        use_ioc_tables=None,
    ):
        """."""
        setpoints, polarization, pparameter_value, kparameter_value = (
            self.calculate_setpoints(
                pparameter_value, kparameter_value, use_ioc_tables
            )
        )
        if dry_run:
            print(f'use_ioc_tables : {use_ioc_tables}')
            print(f'polarization   : {polarization}')
            print(f'pparameter     : {pparameter_value}')
            print(f'kparameter     : {kparameter_value}')
            print()
        devcorrs = []
        devcorrs += self.chdevs
        devcorrs += self.cvdevs
        devcorrs += self.qsdevs
        devcorrs += self.lcdevs
        devcorrs += self.ccdevs
        corrs = dict()
        for devcorr in devcorrs:
            # TODO: check power supply status
            pvname_sp = devcorr.devname + ':Current-SP'
            curr0 = devcorr.current_mon  # after an interlock, RB <> Mon=0
            curr1 = setpoints[pvname_sp]
            corrs[devcorr.devname] = (devcorr, curr0, curr1)

        for idx in range(nrpts):
            delta_ramp = (idx + 1) / nrpts
            if dry_run:
                print(f'point {idx + 1}/{nrpts}')
            for psname in corrs:
                devcorr, curr0, curr1 = corrs[psname]
                curr = curr0 + delta_ramp * (curr1 - curr0)
                if dry_run:
                    print(f'{psname:<20s}: {curr:+.6f}')
                else:
                    devcorr.set_current(curr, wait_mon=True)
            if dry_run:
                print()
            _time.sleep(time_interval / (nrpts - 1))

    def _create_devices(self, props2init_ctrl, props2init_corrs):
        devctrl = (
            None
            if not self._with_devctrl
            else IDFFCtrl(devname=self._devname, props2init=props2init_ctrl)
        )
        pol_mon = _ID.get_idclass(self.iddevname).PARAM_PVS.POL_MON
        params = (self._pparametername, self._kparametername, pol_mon)
        props2init = tuple(param for param in params if param is not None)
        devid = _ID(
            devname=self.iddevname,
            props2init=props2init,
            auto_monitor_mon=False,
        )
        psclass = _PowerSupplyFBP
        p2i = props2init_corrs
        devsch = [psclass(devname=dev, props2init=p2i) for dev in self.chnames]
        devscv = [psclass(devname=dev, props2init=p2i) for dev in self.cvnames]
        devsqs = [psclass(devname=dev, props2init=p2i) for dev in self.qsnames]
        devslc = [psclass(devname=dev, props2init=p2i) for dev in self.lcnames]
        devsqn = [psclass(devname=dev, props2init=p2i) for dev in self.qnnames]
        devscc = [psclass(devname=dev, props2init=p2i) for dev in self.ccnames]
        return devctrl, devid, devsch, devscv, devsqs, devslc, devsqn, devscc

    def _create_labels_2_corrdevs_dict(self):
        ch_labels = self.idffclass.IDFF_CH_LABELS
        cv_labels = self.idffclass.IDFF_CV_LABELS
        qs_labels = self.idffclass.IDFF_QS_LABELS
        lc_labels = self.idffclass.IDFF_LC_LABELS
        qn_labels = self.idffclass.IDFF_QN_LABELS
        cc_labels = self.idffclass.IDFF_CC_LABELS
        devs = dict()
        devs.update({lab: dev for lab, dev in zip(ch_labels, self._devsch)})
        devs.update({lab: dev for lab, dev in zip(cv_labels, self._devscv)})
        devs.update({lab: dev for lab, dev in zip(qs_labels, self._devsqs)})
        devs.update({lab: dev for lab, dev in zip(lc_labels, self._devslc)})
        devs.update({lab: dev for lab, dev in zip(qn_labels, self._devsqn)})
        devs.update({lab: dev for lab, dev in zip(cc_labels, self._devscc)})
        return devs

    def _get_corrnames(self, labels):
        conv = _IDSearch.conv_idname_labels_2_corrnames
        return conv(self.iddevname, labels)



"""Gamma Monitor Search module."""

from ..namesys import Filter as _Filter, SiriusPVName as _SiriusPVName


class GammaMonitorSearch:
    """Gamma Monitor Search Class."""

    # TODO: This mapping should be moved to a configuration file at
    # control-system-constants, and not hardcoded in the code.
    _gamma2counter = {
        'SI-01M2:CO-Gamma': 'SI-01C2:CO-Counter:Ch1',
        'SI-01C1:CO-Gamma': 'SI-01C2:CO-Counter:Ch2',
        'SI-01C2:CO-Gamma': 'SI-01C3:CO-Counter:Ch1',
        'SI-01C3:CO-Gamma': 'SI-01C3:CO-Counter:Ch2',
        'SI-01C4:CO-Gamma': 'SI-02M1:CO-Counter:Ch1',
        'SI-02M2:CO-Gamma': 'SI-02C2:CO-Counter:Ch1',
        'SI-02C1:CO-Gamma': 'SI-02C2:CO-Counter:Ch2',
        'SI-02C2:CO-Gamma': 'SI-02C3:CO-Counter:Ch1',
        'SI-02C3:CO-Gamma': 'SI-02C3:CO-Counter:Ch2',
        'SI-02C4:CO-Gamma': 'SI-03M1:CO-Counter:Ch1',
        'SI-03M2:CO-Gamma': 'SI-03C2:CO-Counter:Ch1',
        'SI-03C1:CO-Gamma': 'SI-03C2:CO-Counter:Ch2',
        'SI-03C2:CO-Gamma': 'SI-03C3:CO-Counter:Ch1',
        'SI-03C3:CO-Gamma': 'SI-03C3:CO-Counter:Ch2',
        'SI-03C4:CO-Gamma': 'SI-04M1:CO-Counter:Ch1',
        'SI-04M2:CO-Gamma': 'SI-04C2:CO-Counter:Ch1',
        'SI-04C1:CO-Gamma': 'SI-04C2:CO-Counter:Ch2',
        'SI-04C2:CO-Gamma': 'SI-04C3:CO-Counter:Ch1',
        'SI-04C3:CO-Gamma': 'SI-04C3:CO-Counter:Ch2',
        'SI-04C4:CO-Gamma': 'SI-05M1:CO-Counter:Ch1',
        'SI-05M2:CO-Gamma': 'SI-05C2:CO-Counter:Ch1',
        'SI-05C1:CO-Gamma': 'SI-05C2:CO-Counter:Ch2',
        'SI-05C2:CO-Gamma': 'SI-05C3:CO-Counter:Ch1',
        'SI-05C3:CO-Gamma': 'SI-05C3:CO-Counter:Ch2',
        'SI-05C4:CO-Gamma': 'SI-06M1:CO-Counter:Ch1',
        'SI-06M2:CO-Gamma': 'SI-06C2:CO-Counter:Ch1',
        'SI-06C1:CO-Gamma': 'SI-06C2:CO-Counter:Ch2',
        'SI-06C2:CO-Gamma': 'SI-06C3:CO-Counter:Ch1',
        'SI-06C3:CO-Gamma': 'SI-06C3:CO-Counter:Ch2',
        'SI-06C4:CO-Gamma': 'SI-07M1:CO-Counter:Ch1',
        'SI-07M2:CO-Gamma': 'SI-07C2:CO-Counter:Ch1',
        'SI-07C1:CO-Gamma': 'SI-07C2:CO-Counter:Ch2',
        'SI-07C2:CO-Gamma': 'SI-07C3:CO-Counter:Ch1',
        'SI-07C3:CO-Gamma': 'SI-07C3:CO-Counter:Ch2',
        'SI-07C4:CO-Gamma': 'SI-08M1:CO-Counter:Ch1',
        'SI-08M2:CO-Gamma': 'SI-08C2:CO-Counter:Ch1',
        'SI-08C1:CO-Gamma': 'SI-08C2:CO-Counter:Ch2',
        'SI-08C2:CO-Gamma': 'SI-08C3:CO-Counter:Ch1',
        'SI-08C3:CO-Gamma': 'SI-08C3:CO-Counter:Ch2',
        'SI-08C4:CO-Gamma': 'SI-09M1:CO-Counter:Ch1',
        'SI-09M2:CO-Gamma': 'SI-09C2:CO-Counter:Ch1',
        'SI-09C1:CO-Gamma': 'SI-09C2:CO-Counter:Ch2',
        'SI-09C2:CO-Gamma': 'SI-09C3:CO-Counter:Ch1',
        'SI-09C3:CO-Gamma': 'SI-09C3:CO-Counter:Ch2',
        'SI-09C4:CO-Gamma': 'SI-10M1:CO-Counter:Ch1',
        'SI-10M2:CO-Gamma': 'SI-10C2:CO-Counter:Ch1',
        'SI-10C1:CO-Gamma': 'SI-10C2:CO-Counter:Ch2',
        'SI-10C2:CO-Gamma': 'SI-10C3:CO-Counter:Ch1',
        'SI-10C3:CO-Gamma': 'SI-10C3:CO-Counter:Ch2',
        'SI-10C4:CO-Gamma': 'SI-11M1:CO-Counter:Ch1',
        'SI-11M2:CO-Gamma': 'SI-11C2:CO-Counter:Ch1',
        'SI-11C1:CO-Gamma': 'SI-11C2:CO-Counter:Ch2',
        'SI-11C2:CO-Gamma': 'SI-11C3:CO-Counter:Ch1',
        'SI-11C3:CO-Gamma': 'SI-11C3:CO-Counter:Ch2',
        'SI-11C4:CO-Gamma': 'SI-12M1:CO-Counter:Ch1',
        'SI-12M2:CO-Gamma': 'SI-12C2:CO-Counter:Ch1',
        'SI-12C1:CO-Gamma': 'SI-12C2:CO-Counter:Ch2',
        'SI-12C2:CO-Gamma': 'SI-12C3:CO-Counter:Ch1',
        'SI-12C3:CO-Gamma': 'SI-12C3:CO-Counter:Ch2',
        'SI-12C4:CO-Gamma': 'SI-13M1:CO-Counter:Ch1',
        'SI-13M2:CO-Gamma': 'SI-13C2:CO-Counter:Ch1',
        'SI-13C1:CO-Gamma': 'SI-13C2:CO-Counter:Ch2',
        'SI-13C2:CO-Gamma': 'SI-13C3:CO-Counter:Ch1',
        'SI-13C3:CO-Gamma': 'SI-13C3:CO-Counter:Ch2',
        'SI-13C4:CO-Gamma': 'SI-14M1:CO-Counter:Ch1',
        'SI-14M2:CO-Gamma': 'SI-14C2:CO-Counter:Ch1',
        'SI-14C1:CO-Gamma': 'SI-14C2:CO-Counter:Ch2',
        'SI-14C2:CO-Gamma': 'SI-14C3:CO-Counter:Ch1',
        'SI-14C3:CO-Gamma': 'SI-14C3:CO-Counter:Ch2',
        'SI-14C4:CO-Gamma': 'SI-15M1:CO-Counter:Ch1',
        'SI-15M2:CO-Gamma': 'SI-15C2:CO-Counter:Ch1',
        'SI-15C1:CO-Gamma': 'SI-15C2:CO-Counter:Ch2',
        'SI-15C2:CO-Gamma': 'SI-15C3:CO-Counter:Ch1',
        'SI-15C3:CO-Gamma': 'SI-15C3:CO-Counter:Ch2',
        'SI-15C4:CO-Gamma': 'SI-16M1:CO-Counter:Ch1',
        'SI-16M2:CO-Gamma': 'SI-16C2:CO-Counter:Ch1',
        'SI-16C1:CO-Gamma': 'SI-16C2:CO-Counter:Ch2',
        'SI-16C2:CO-Gamma': 'SI-16C3:CO-Counter:Ch1',
        'SI-16C3:CO-Gamma': 'SI-16C3:CO-Counter:Ch2',
        'SI-16C4:CO-Gamma': 'SI-17M1:CO-Counter:Ch1',
        'SI-17M2:CO-Gamma': 'SI-17C2:CO-Counter:Ch1',
        'SI-17C1:CO-Gamma': 'SI-17C2:CO-Counter:Ch2',
        'SI-17C2:CO-Gamma': 'SI-17C3:CO-Counter:Ch1',
        'SI-17C3:CO-Gamma': 'SI-17C3:CO-Counter:Ch2',
        'SI-17C4:CO-Gamma': 'SI-18M1:CO-Counter:Ch1',
        'SI-18M2:CO-Gamma': 'SI-18C2:CO-Counter:Ch1',
        'SI-18C1:CO-Gamma': 'SI-18C2:CO-Counter:Ch2',
        'SI-18C2:CO-Gamma': 'SI-18C3:CO-Counter:Ch1',
        'SI-18C3:CO-Gamma': 'SI-18C3:CO-Counter:Ch2',
        'SI-18C4:CO-Gamma': 'SI-19M1:CO-Counter:Ch1',
        'SI-19M2:CO-Gamma': 'SI-19C2:CO-Counter:Ch1',
        'SI-19C1:CO-Gamma': 'SI-19C2:CO-Counter:Ch2',
        'SI-19C2:CO-Gamma': 'SI-19C3:CO-Counter:Ch1',
        'SI-19C3:CO-Gamma': 'SI-19C3:CO-Counter:Ch2',
        'SI-19C4:CO-Gamma': 'SI-20M1:CO-Counter:Ch1',
        'SI-20M2:CO-Gamma': 'SI-20C2:CO-Counter:Ch1',
        'SI-20C1:CO-Gamma': 'SI-20C2:CO-Counter:Ch2',
        'SI-20C2:CO-Gamma': 'SI-20C3:CO-Counter:Ch1',
        'SI-20C3:CO-Gamma': 'SI-20C3:CO-Counter:Ch2',
        'SI-20C4:CO-Gamma': 'SI-01M1:CO-Counter:Ch1',
    }
    _gamma2counter = {
        _SiriusPVName(gamma): _SiriusPVName(counter)
        for gamma, counter in _gamma2counter.items()
    }

    @staticmethod
    def get_gammanames(filters=None):
        """Return a sorted and filtered list of all Gamma names."""
        gamma_names_list = list(GammaMonitorSearch._gamma2counter.keys())
        gamma_names = _Filter.process_filters(
            gamma_names_list, filters=filters
        )
        return sorted(gamma_names)

    @staticmethod
    def get_counter_names(filters=None):
        """Return a sorted and filtered list of all Counter names."""
        counter_names = list({
            dev.device_name
            for dev in GammaMonitorSearch._gamma2counter.values()
        })
        counter_names = _Filter.process_filters(counter_names, filters=filters)
        return sorted(counter_names)

    @staticmethod
    def conv_gammaname_2_countername(gammaname):
        """Return the counter name to which the gamma monitor is connected."""
        return GammaMonitorSearch._gamma2counter[gammaname].device_name

    @staticmethod
    def conv_gammaname_2_counter_channel(gammaname):
        """Return the counter channel of a Gamma Monitor."""
        return GammaMonitorSearch._gamma2counter[gammaname].propty
